/****************************************
 *  COPYRIGHT (C) 2012
 *  Holger Graf
 ****************************************/
package sitool.components;

import java.awt.Point;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author graf
 *
 */
public class Coordinates
{
    public final static int sysScale = 50;
    public final static int galScale = sysScale * sysScale;
    public static final long scale = 15000000;
    
    private static int ringStart[] = { 1, 7, 19, 39, 67, 99, 139, 187 };

    private int x = -1;
    private int y = -1;
    private int orbit = -1;

    // the position this Coordinates map to
    private Position position = null;

    private static Pattern pattern = Pattern.compile( "(\\[?(\\d+:\\d+:\\d+)\\]?\\s+(\\S\\s+)?)?(\\d+):(\\d+)\\|(\\d+):(\\d+)\\|(\\d+):(\\d+)\\|(\\d+)" );
    //private static Pattern pattern = Pattern.compile( "(\\d+):(\\d+)\\|(\\d+):(\\d+)\\|(\\d+):(\\d+)\\|(\\d+)" );

    public Coordinates( String coordinate )
    {
        Matcher m = pattern.matcher( coordinate );
        if( m.find() )
        {
            Point galaxy = new Point( Integer.valueOf( m.group( 4 ) ), Integer.valueOf( m.group( 5 ) ) );
            Point system = new Point( Integer.valueOf( m.group( 6 ) ), Integer.valueOf( m.group( 7 ) ) );
            Point planet = new Point( Integer.valueOf( m.group( 8 ) ), Integer.valueOf( m.group( 9 ) ) );
            orbit = Integer.valueOf( m.group( 10 ) );
            if( m.group( 1 ) != null )
            {
                position = new Position( m.group( 2 ) );
            }
            init( galaxy, system, planet, orbit );
        }
    }

    public Coordinates( Point galaxy, Point system, Point planet, int orbit )
    {
        init( galaxy, system, planet, orbit );
    }

    public Coordinates(Position position, int start, int orbit)
    {
        // these are the 'normalized' coordinates, set the tick values to 0
        Point galaxy = getGalaxyPoint( position, 0 );
        Point system = getSystemPoint( position );
        Point planet = getPlanetStartPoint( orbit, start );
        init( galaxy, system, planet, orbit );
        this.position = position;
    }

    /**
     * This constructor constructs Coordinates at the absolute position. They are probably invalid
     * @param x
     * @param y
     */
    public Coordinates( int x, int y )
    {
        this.x = x;
        this.y = y;
    }

    private void init( Point galaxy, Point system, Point planet, int orbit )
    {
        x = galScale * ( galaxy.x - 1 );
        y = galScale * ( galaxy.y - 1 );
        x += sysScale * ( system.x - 1 );
        y += sysScale * ( system.y - 1 );
        x += ( planet.x - 1 );
        y += ( planet.y - 1 );
        this.orbit = orbit;
    }

    public static Point getGalaxyPoint( Position position, int galaxyTicks )
    {
        // get the galaxy position
        int galaxy = position.getGalaxy();
        int ring = getRing( galaxy );
        Point[] galaxies = Universe.getRing( ring );
        int point = ( galaxy - ringStart[ring] + galaxyTicks ) % galaxies.length;
        if( point < 0 ) point += galaxies.length;
        
        Point galaxyXy = galaxies[point];
        return galaxyXy;
    }

    public static int getGalaxy( Point galaxyPoint, int galaxyTicks )
    {
        int rv = -1;
        // is there such a galaxy point?
        int ring = getRing(galaxyPoint);
        if( ring >= 0)
        {
            List<Point> galaxies = Universe.getRingList( ring );
            int point = galaxies.indexOf( galaxyPoint );
            if( point >= 0 )
            {
                // translate back
                point = ( point - galaxyTicks ) % galaxies.size();
                if( point < 0 ) point += galaxies.size();
    
                if( ring < ringStart.length)
                {
                    rv = point + ringStart[ring];
                }
            }
        }
        return rv;
    }

    public static int getSystem( int galaxy, Point systemPoint )
    {
        int rv = -1;
        try
        {
            List<Point> systems = Galaxies.getSystemsList( galaxy - 1 );
            int point = systems.indexOf( systemPoint );
            if( point != -1 )
            {
                rv = point + 1;
            }
        }
        catch(IndexOutOfBoundsException ex) 
        {
            // trying to access a galaxy that does not (yet) exist
        }
        return rv;
    }

    public static Point getSystemPoint( Position position )
    {
        int galaxy = position.getGalaxy();
        // system position is thankfully fixed
        Point systemXy = Galaxies.getSystems( galaxy - 1 )[position.getSystem() - 1];
        return systemXy;
    }

    public static Point getPlanetStartPoint( int orbit, int start )
    {
        Point planetXy = Orbits.getOrbit( orbit )[start];
        return planetXy;
    }

    public long distance( Coordinates other )
    {
        return Math.round( scale * unitDistance( other ) );
    }

    public double unitDistance( Coordinates other )
    {
        int dx = x - other.x;
        int dy = y - other.y;
        return Math.sqrt( dx * dx + dy * dy );
    }

    public Position getPosition()
    {
        return position;
    }

    public void setPosition( Position position )
    {
        this.position = position;
    }

    public Point getGalaxy()
    {
        return new Point( x / galScale + 1, y / galScale + 1 );
    }

    public Point getSystem()
    {
        return new Point( ( x % galScale ) / sysScale + 1, ( y % galScale ) / sysScale + 1 );
    }

    public Point getPlanet()
    {
        return new Point( x % 50 + 1, y % 50 + 1 );
    }

    public int getX() {
        return x;
    }

    public int getY() {
        return y;
    }

    public int getOrbit()
    {
        return orbit;
    }

    /**
     * Return the index on the orbit the current coordinates represent.
     * <code>-1</code> if no match is found.
     * @return
     */
    public int getOrbitPosition()
    {
        int pos = -1;
        if( orbit != -1 )
        {
            List<Point> positions = Orbits.getOrbitList( orbit );
            pos = positions.indexOf( getPlanet() );
        }
        return pos;
    }

    public boolean isValid()
    {
        return (x >= 0 && y >= 0 && orbit >= 0);
    }

//    public Coordinates translate( int galaxyTicks, int ticks )
//    {
//        Coordinates translatedCoordinates = this;
//        if( isValid() )
//        {
//            Point galaxy = getGalaxyPoint( position, galaxyTicks );
//            Point system = getSystem();
//            Point planet = translatePlanetPoint( ticks );
//            translatedCoordinates = new Coordinates( galaxy, system, planet, orbit );
//            translatedCoordinates.setPosition( position );
//        }
//        return translatedCoordinates;
//    }
//
    /**
     * shift coordinates to the next or previous orbital position, depending on the sign of ticks
     * @param ticks positive - shift forward, negative - shift backwards
     */
    public void shift( int ticks )
    {
        if( isValid() )
        {
            int move = (ticks > 0) ? 1 : (ticks < 0) ? -1 : 0;
            Point planet = translatePlanetPoint( move );
            x = (x / sysScale) * sysScale + planet.x - 1;
            y = (y / sysScale) * sysScale + planet.y - 1;
        }
    }

    private Point translatePlanetPoint( int ticks )
    {
        List<Point> orbit = Orbits.getOrbitList( this.orbit );
        Point planetXy = getPlanet();
        int index= orbit.indexOf( planetXy );
        if( index != -1 )
        {
            int point = ( index + ticks ) % orbit.size();
            if( point < 0 ) point += orbit.size();
            planetXy = Orbits.getOrbit( this.orbit )[point];
        }
        return planetXy;
    }

    private static int getRing( int galaxy  )
    {
        int ring = 7;
        
        while( ring > 0 && galaxy < ringStart[ring])
        {
            ring--;
        }
        return ring;
    }

    private static int getRing( Point galaxy  )
    {
        int ring = Universe.getUniverseList().indexOf( galaxy );
        if( ring >= 0)
        {
            ring = getRing( ring + 1 );
        }
        return ring;
    }

    @Override
    public int hashCode() {
        return 1000000 * x + y;
    }

    @Override
    public boolean equals(Object obj) {
        boolean rv = false;
        if( obj instanceof Coordinates )
        {
            Coordinates that = (Coordinates)obj;
            rv = this.x == that.x && this.y == that.y;
        }
        return rv;
    }

    public String toFullString()
    {
        StringBuilder sb = new StringBuilder();
        
        if( position.isRealPosition() )
        {
            sb.append( position.toString() );
            sb.append( " - " );
        }
        sb.append( toString() );
        return sb.toString();
    }
    
    public String toString()
    {
        StringBuilder sb = new StringBuilder();

        if( isValid() )
        {
            int dx = x % 2500;
            int dy = y % 2500;
            sb.append( x / 2500 + 1).append( ':' );
            sb.append( y / 2500 + 1).append( '|' );
            sb.append( dx / 50  + 1).append( ':' );
            sb.append( dy / 50 + 1).append( '|' );
            sb.append( dx % 50 + 1).append( ':' );
            sb.append( dy % 50 + 1).append( '|' );
            sb.append( orbit );
        }

        return sb.toString();
    }
}
