/****************************************
 *  COPYRIGHT (C) 2021
 *  Holger Graf
 ****************************************/
package sitool.datatransfer;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import sitool.Account;
import sitool.components.Building;
import sitool.components.Defense;
import sitool.components.Fleet;
import sitool.components.Planet;
import sitool.components.Position;
import sitool.components.Research;
import sitool.components.Ship;
import sitool.components.Specialization;
import sitool.components.Stage;
import sitool.components.Technology;
import sitool.components.Version;
import sitool.json.IntergalacticOverviewBean;
import sitool.json.LevelBean;
import sitool.json.PlanetBean;
import sitool.json.PlanetType;
import sitool.json.SpecsBean;
import sitool.json.SpyReportBean;
import sitool.json.SuperComputerBean;
import sitool.resources.ResourcePane;

/**
 * @author graf
 *
 */
public class DataImport {

    public DataImport() {}

    public static List<Planet> createPlanetsFromJSON(Version version, IntergalacticOverviewBean overview, boolean isOutpost, boolean includeBuildings, boolean includeFleet) {
        final List<Planet> planets = new ArrayList<>();
        for(PlanetBean pb : overview.getPlanets()) {
            final Position position = new Position(pb.getCoords());
            if(!(isOutpost ^ position.getGalaxy() == version.getOutpostGalaxy())) {
                final Planet planet = new Planet( pb.getName(), position );
                planet.setMine(true);
                planets.add(planet);
                planet.setSize(pb.getFields());
                fillPlanetFromJSON(version, planet, pb, false, includeBuildings, includeFleet);
                // asteroid?
                if(pb.getAstro() != null) {
                    PlanetBean ab = pb.getAstro();
                    planet.setAstroSize(ab.getFields());
                    fillPlanetFromJSON(version, planet, ab, true, includeBuildings, includeFleet);
                }
            }
        }
        return planets;
    }

    public static Planet createPlanetFromJSON(Version version, SpyReportBean reportBean, boolean includeBuildings, boolean includeFleet) {
        PlanetBean pl = reportBean.getPlanet();
        final String coords = pl.getCoords();
        final Position position = new Position(coords);
        final boolean astro = (pl.getType() == PlanetType.MOON);
        position.setAsteroidPosition(astro);
        final String name = pl.getName();
        final Planet planet = new Planet( name, position );
        planet.setSize( 1000 );
        planet.setAstroSize( astro ? 100 : 0 );

        fillPlanetFromJSON(version, planet, pl, astro, includeBuildings, includeFleet);
        return planet;
    }

    public static Technology createTechnologyFromJSON(Map<String, LevelBean> researches) {
        Technology technology = null;
        if(researches != null) {
            technology = new Technology();
            for( Map.Entry<String, LevelBean> entry : researches.entrySet() ) {
                Research research = Research.fromId(entry.getKey());
                technology.setTech( research, (int)entry.getValue().getLevel() );
            }
        }
        return technology;
    }


    public static Specialization createSpecializationFromJSON( SpecsBean specs ) {
        Specialization specialization = new Specialization();
        specialization.setFleeter(specs.getFleeter());
        specialization.setMiner(specs.getMiner());
        specialization.setResearcher(specs.getResearcher());
        return specialization;
    }

    public static void updateConfiguration( Account account, List<Planet> planets, Technology newTech, Specialization specs, Fleet totalShips )
    {
        if(account != null) {
            boolean isOutpost = account.isOutpost();
            Version version = account.getVersion();
            // planets
            if(planets != null) {
                List<Planet> myPlanets = account.getMyPlanets();
                for( Planet planet : planets) {
                    if(!(isOutpost ^ planet.getPosition().getGalaxy() == version.getOutpostGalaxy())) {
                        final int index = myPlanets.indexOf(planet);
                        if( index < 0) {
                            Planet newPlanet = new Planet(planet.getName(), planet.getPosition());
                            newPlanet.setMine(true);
                            newPlanet.update(planet);
                            myPlanets.add(newPlanet);
                        }
                        else {
                            myPlanets.get(index).update(planet);
                        }
                    }
                }
            }
            // total ships
            if(totalShips != null) {
                account.setTotalShips(totalShips);
            }
            // tech
            if( newTech != null && newTech.isValid() ) {
                account.setMyTech( newTech );
            }
            // specialization
            if(specs != null) {
                Specialization specialization = account.getSpecialization();
                specialization.setFleeter(specs.getFleeter());
                specialization.setMiner(specs.getMiner());
                specialization.setResearcher(specs.getResearcher());
            }
        }
    }

    public static Date importResourcesFromJSON(ResourcePane resourcePane, IntergalacticOverviewBean overview) {
        for(PlanetBean pb : overview.getPlanets()) {
            final Position position = new Position(pb.getCoords());
            final Planet planet = new Planet( pb.getName(), position );
            List<Long> resources = pb.getResources();
            resourcePane.updateStock(planet, resources);
            if(pb.getAstro() != null) {
                PlanetBean ab = pb.getAstro();
                position.setAsteroidPosition(true);
                final Planet asteroid = new Planet( pb.getName(), position );
                resources = ab.getResources();
                resourcePane.updateStock(asteroid, resources);
            }
            resourcePane.updateStock(overview.getResources().getOnMission());
        }
        return new Date(overview.getInfo().getTs());
    }

    public static Date importResourcesFromJSON(ResourcePane resourcePane, SpyReportBean reportBean) {
        final String coords = reportBean.getPlanet().getCoords();
        final Position position = new Position(coords);
        final boolean astro = (reportBean.getPlanet().getType() == PlanetType.MOON);
        position.setAsteroidPosition(astro);
        final String name = reportBean.getPlanet().getName();
        final Planet planet = new Planet( name, position );
        final List<Long> resources = reportBean.getPlanet().getResources();
        resourcePane.updateStock(planet, resources);
        return new Date(reportBean.getInfo().getTs());
    }

    private static void fillPlanetFromJSON(Version version, Planet planet, PlanetBean pl, boolean isAsteroid, boolean includeBuildings, boolean includeFleet) {
        // buildings
        Map<String, LevelBean> buildings = pl.getBuildings();
        if(includeBuildings && buildings != null) {
            for( Map.Entry<String, LevelBean> entry : buildings.entrySet() ) {
                Building building = Building.fromId(entry.getKey());
                Stage stage = new Stage(building, (int)entry.getValue().getLevel());
                planet.addBuilding(building, stage, isAsteroid);
            }
            // super computer
            SuperComputerBean sbean = pl.getScomputer();
            if(!isAsteroid && sbean != null) {
                Building building = Building.superComputer;
                Stage stage = new Stage(building, sbean.getLevel_raw());
                planet.addBuilding(building, stage, isAsteroid);
            }
        }
        else {
            // mark buildings invalid
            planet.resetBuildings(isAsteroid);
        }
        if(includeFleet) {
            Fleet fleet = planet.getFleet(isAsteroid);
            // defenses
            Map<String, LevelBean> defenses = pl.getDefense();
            if(defenses != null) {
                for( Map.Entry<String, LevelBean> entry : defenses.entrySet() ) {
                    Defense defense = Defense.fromId(entry.getKey());
                    fleet.setCount( defense,  entry.getValue().getLevel() );
                }
            }
            else {
                fleet.resetDefenses();
            }
            // ships
            Map<String, LevelBean> ships = pl.getShips();
            if(ships != null) {
                for( Map.Entry<String, LevelBean> entry : ships.entrySet() ) {
                    Ship ship = Ship.fromId(entry.getKey());
                    if(ship != null)
                        fleet.setCount( ship, entry.getValue().getLevel() );
                }
            }
            else {
                fleet.resetShips();
            }
        }
        else {
            planet.resetFleet(isAsteroid);
        }
    }

    public static Fleet importTotalShipsFromJSON( Map<String, LevelBean> ships, boolean isOutpost ) {
        Fleet rv = null;
        if( ships != null && !isOutpost) {
            rv = new Fleet();
            for(Map.Entry<String, LevelBean> entry : ships.entrySet()) {
                Ship ship = Ship.fromId(entry.getKey());
                if(ship != null)
                    rv.setCount( ship, entry.getValue().getLevel() );
            }
        }
        return rv;
    }

}
