/****************************************
 *  COPYRIGHT (C) 2021
 *  Holger Graf
 ****************************************/
package sitool.fleet;

import java.util.EnumSet;

import sitool.components.Fleet;
import sitool.components.Ship;

/**
 * @author graf
 *
 */
public class TacticalCalculator {

    public final static EnumSet<Ship> ships = EnumSet.of(
                    Ship.smallTransport,
                    Ship.largeTransport,
                    Ship.transmitter,
                    Ship.smallFighter,
                    Ship.manOWar,
                    Ship.starCruiser,
                    Ship.phoenix,
                    Ship.colonyShip,
                    Ship.recycler,
                    Ship.battleship,
                    Ship.stealthBomber,
                    Ship.destroyer,
                    Ship.imperialStarbase
                 );

    private static enum BonusRelation {
        SmallFighterBattleShip(Ship.smallFighter, Ship.battleship, 1, 0.1),
        SmallFighterDestroyer(Ship.smallFighter, Ship.destroyer, 30, 2),
        SmallFighterStarBase(Ship.smallFighter, Ship.imperialStarbase, 10000, 3),
        ManOWarBattleShip(Ship.manOWar, Ship.battleship, 1, 0.1),
        ManOWarStealthBomber(Ship.manOWar, Ship.stealthBomber, 100, 4),
        ManOWarDestroyer(Ship.manOWar, Ship.destroyer, 10, 2),
        StarCruiserBattleShip(Ship.starCruiser, Ship.battleship, 1, 0.1),
        StarCruiserStealthBomber(Ship.starCruiser, Ship.stealthBomber, 50, 3),
        StarCruiserDestroyer(Ship.starCruiser, Ship.destroyer, 10, 2),
        PhoenixBattleShip(Ship.phoenix, Ship.battleship, 1, 0.1),
        PhoenixDestroyer(Ship.phoenix, Ship.destroyer, 5, 2),
        PhoenixStarBase(Ship.phoenix, Ship.imperialStarbase, 100, 2),
        StealthBomberBattleShip(Ship.stealthBomber, Ship.battleship, 1, 0.1),
        DestroyerBattleShip(Ship.destroyer, Ship.battleship, 1, 0.1),
        StarBaseBattleShip(Ship.imperialStarbase, Ship.battleship, 0.4, 0.1),
        ;

        private final Ship one;
        private final Ship two;
        private final double relation;
        private final double multiplier;

        private BonusRelation(Ship one, Ship two, double relation, double multiplier) {
            this.one = one;
            this.two = two;
            this.relation = relation;
            this.multiplier = multiplier;
        }

        public long getBonus(Fleet fleet) {
            long countOne = fleet.getCount(one);
            long countTwo = fleet.getCount(two);
            double factor = Math.min(countOne, countTwo * relation);
            return (long)(factor * multiplier * one.getFirePower());
        }

        public static EnumSet<BonusRelation> bonusRelationByShip(final Ship ship) {
            EnumSet<BonusRelation> rv;
            switch(ship) {
                case smallFighter:
                    rv = EnumSet.of(SmallFighterBattleShip,
                                    SmallFighterDestroyer,
                                    SmallFighterStarBase);
                    break;
                case manOWar:
                    rv = EnumSet.of(ManOWarBattleShip,
                                    ManOWarStealthBomber,
                                    ManOWarDestroyer);
                    break;
                case starCruiser:
                    rv = EnumSet.of(StarCruiserBattleShip,
                                    StarCruiserStealthBomber,
                                    StarCruiserDestroyer);
                    break;
                case phoenix:
                    rv = EnumSet.of(PhoenixBattleShip,
                                    PhoenixDestroyer,
                                    PhoenixStarBase);
                    break;
                case stealthBomber:
                    rv = EnumSet.of(StealthBomberBattleShip);
                    break;
                case destroyer:
                    rv = EnumSet.of(DestroyerBattleShip);
                    break;
                case imperialStarbase:
                    rv = EnumSet.of(StarBaseBattleShip);
                    break;
                default:
                    rv = EnumSet.noneOf(BonusRelation.class);
                    break;
            }
            return rv;
        }
    }


    public static long tacticalStrength(final Fleet fleet) {
        long rv = 0;
        if(fleet.hasShips()) {
            rv = baseStrength(fleet);
            rv += bonusStrength(fleet);
        }
        return rv;
    }

    public static long baseStrength(final Ship ship, final Fleet fleet) {
        long rv = 0;
        long count = fleet.getCount(ship);
        if(count > 0) {
            rv = Math.round(shipStrength(ship) * count * baseFactor(ship));
        }
        return rv;
    }


    public static long baseStrength(final Fleet fleet) {
        double rv = 0;
        for(Ship ship : ships) {
            long count = fleet.getCount(ship);
            if(count > 0) {
                rv +=  shipStrength(ship) * count * baseFactor(ship);
            }
        }
        return Math.round(rv);
    }


    public static long bonusStrength( final Ship ship, final Fleet fleet ) {
        long rv = 0;
        long count = fleet.getCount(ship);
        if(count > 0) {
            for(BonusRelation bonusRelation : BonusRelation.bonusRelationByShip(ship)) {
                rv += bonusRelation.getBonus(fleet);
            }
        }
        return rv;
    }

    public static long bonusStrength( final Fleet fleet ) {
        long rv = 0;
        for(Ship ship : ships) {
            rv += bonusStrength(ship, fleet);
        }
        return rv;
    }

    /**
     * $iBase = ($aCfg["damage"] + ($aCfg["shield"]*0.1)) * $iAmount *
     *          (IN($iShipID, 1, 2, 3) ? 0.3 :
     *          (IN($iShipID, 8, 11) ? 0.5 : 1));
     */
    private static double shipStrength(final Ship ship) {
        return ship.getFirePower() + 0.1 * ship.getShields();
    }

    private static double baseFactor(final Ship ship) {
        return (ship == Ship.smallTransport || ship == Ship.largeTransport || ship == Ship.transmitter )
               ? 0.3 : (ship == Ship.colonyShip || ship == Ship.recycler )
               ? 0.5 : 1.0;
    }
}
