/****************************************
 *  COPYRIGHT (C) 2022
 *  Holger Graf
 ****************************************/
package sitool.planets;

import java.awt.Point;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.jr.ob.JSON;

import sitool.ConfigData;
import sitool.components.Coordinates;
import sitool.components.Orbits;
import sitool.components.Position;
import sitool.json.CoordinatesBean;
import sitool.json.CoordinatorBean;
import sitool.json.CurrentTicksBean;

/**
 * @author graf
 *
 */
public class Coordinator {

    private static String upLink = "https://spaceinvasion2.bitmeup.com/api/planet/sync/";

    private int currentGalaxyTicks = 0;
    private int currentTicks = 0;

    private final ConfigData configData;

    protected class CoordinatesContainer {

        private final Position position;
        private final Coordinates baseCoordinates;
        private final Coordinates currentCoordinates;

        public CoordinatesContainer(CoordinatesBean bean) {
            this.position = bean.getCoord();
            this.baseCoordinates = configData.getCoordinates().get(this.position);
            this.currentCoordinates = new Coordinates(bean.getG(), bean.getS(), bean.getP(), this.baseCoordinates.getOrbit());
        }

        public Position getPosition() {
            return position;
        }

        public Coordinates getBaseCoordinates() {
            return baseCoordinates;
        }

        public Coordinates getCurrentCoordinates() {
            return currentCoordinates;
        }
    }

    public Coordinator(ConfigData configData, boolean noVerify) throws IOException {
        this.configData = configData;
        final URL url = new URL(upLink);
        try (BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()))) {
            int c = 0;
            StringBuilder json = new StringBuilder();
            while ((c = in.read()) != -1) {
                json.append((char) c);
            }
            final CoordinatorBean coordinatorBean = JSON.std.beanFrom(CoordinatorBean.class, json.toString());
            final CurrentTicksBean currentTicksBean = coordinatorBean.getMoves();
            int galaxyTicks = currentTicksBean.getGalaxies();
            int ticks = currentTicksBean.getPlanets();

            if(noVerify) {
                this.currentGalaxyTicks = galaxyTicks;
                this.currentTicks = ticks;
            }
            else {
                final List<CoordinatesBean> coordinatesBeans = coordinatorBean.getPlanets();
                final CoordinatesContainer[] coordinatesContainer = new CoordinatesContainer[coordinatesBeans.size()];
                for(int i = 0; i < coordinatesBeans.size(); i++) {
                    coordinatesContainer[i] = new CoordinatesContainer(coordinatesBeans.get(i));
                }
                if(check(coordinatesContainer, galaxyTicks, ticks)) {
                    this.currentGalaxyTicks = galaxyTicks;
                    this.currentTicks = ticks;
                }
                else {
                    calculate(coordinatesContainer);
                }
            }
        }
    }

    public int getCurrentGalaxyTicks() {
        return currentGalaxyTicks;
    }

    public int getCurrentTicks() {
        return currentTicks;
    }

    private void calculate(CoordinatesContainer[] coordinatesContainer) {
        // galaxy first
        Point current = coordinatesContainer[0].getCurrentCoordinates().getGalaxy();
        while( !Coordinates.getGalaxyPoint( coordinatesContainer[0].getPosition(), currentGalaxyTicks ).equals( current ) )
            currentGalaxyTicks++;

        // planets
        rotate(coordinatesContainer, currentGalaxyTicks);
    }

    private void rotate(CoordinatesContainer[] coordinatesContainers, int galaxyTicks) {
        // cut off switch
        final int orbitLengths[] = new int[coordinatesContainers.length];
        for(int i = 0; i < coordinatesContainers.length; i++) {
            orbitLengths[i] = Orbits.getOrbit(coordinatesContainers[i].getBaseCoordinates().getOrbit()).length;
        }
        final int longest = kgV( orbitLengths );
        final Map<Position, Coordinates> controlMap = buildControlMap(coordinatesContainers, galaxyTicks);
        while(++currentTicks < longest) {
            // shift the systems forward and check coordinates
            configData.shift(1, controlMap);
            boolean compares = true;
            for( CoordinatesContainer coordinatesContainer : coordinatesContainers) {
                Position position = coordinatesContainer.getPosition();
                Coordinates shiftedCoordinates = controlMap.get(position);
                compares &= coordinatesContainer.getCurrentCoordinates().equals(shiftedCoordinates);
            }
            // all matches, done
            if(compares)
                break;
        }
    }

    /**
     * Builds a map of all systems for the planets in <code>coordinatesContainer</code> on day 0
     * with the galaxy already shifted to <code>galaxyTicks</code>
     *
     * @param coordinatesContainer
     * @param galaxyTicks
     *
     * @return map
     */
    private Map<Position, Coordinates> buildControlMap(CoordinatesContainer[] coordinatesContainer, int galaxyTicks) {
        final Map<Position, Coordinates> controlMap = new HashMap<Position, Coordinates>();
        for(CoordinatesContainer container: coordinatesContainer) {
            Position position = container.getPosition();
            // shift the galaxy first
            Point galaxyPoint = Coordinates.getGalaxyPoint(position, galaxyTicks);
            Point systemPoint = Coordinates.getSystemPoint(position);

            // get the 0 day map
            Map<Position, Coordinates> zeroDayMap = configData.getCoordinates();

            int galaxy = position.getGalaxy();
            int system = position.getSystem();
            // add individual planets
            for( int planet = 1; planet <= Position.MAX_PLANETS; planet++ )
            {
                Position key = new Position( galaxy, system, planet );
                Coordinates coordinates = zeroDayMap.get( key );
                // end of system, get out
                if( coordinates == null )
                    break;

                coordinates = new Coordinates( galaxyPoint, systemPoint, coordinates.getPlanet(), coordinates.getOrbit() );
                coordinates.setPosition( key );
                controlMap.put( key, coordinates );
            }
        }
        return controlMap;
    }

    private boolean check(CoordinatesContainer[] coordinatesContainer, int galaxyTicks, int ticks)
    {
        boolean rv = true;
        for( int i = 0; rv && i < coordinatesContainer.length; i++ ) {
            rv &= configData.translate(coordinatesContainer[i].baseCoordinates, galaxyTicks, ticks, false )
                     .equals(coordinatesContainer[i].currentCoordinates);
        }
        return rv;
    }


    private int ggT(int a, int b) {
        int t;
        do {
            t = a % b;
            a = b;
            b = t;
        } while (b != 0);

        return a;
    }

    private int kgV(int a, int b) {
        // ggT(a, b) * kgV(a, b) = a*b
        if (a == 0 || b == 0) return 0;
        return a * b / ggT(a, b);
    }

    private int kgV(int[] is) {
        int rv = 0;
        if( is.length > 0) {
            rv = is[0];
            if (is.length > 1) {
                for (int i = 1; i < is.length; i++) {
                    rv = kgV(rv, is[i]);
                }
            }
        }
        return rv;
    }

}
