/****************************************
 *  COPYRIGHT (C) 2008-2020
 *  Holger Graf
 ****************************************/
package sitool.savecalc;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import sitool.ConfigData;
import sitool.components.Accountable;
import sitool.components.Coordinates;
import sitool.components.Distance;
import sitool.components.FleetSpeed;
import sitool.components.Planet;
import sitool.components.Position;
import sitool.components.PositionSet;
import sitool.components.SystemData;
import sitool.components.Version;

public class Calculator
{
    public static int secondsPerDay = 86400;
    public static int secondsPerHour = 3600;
    public static int secondsPerMinute = 60;
    public static int interval = ( int )( 1.05 * secondsPerHour );

    private ConfigData configData;


    /**
     * @param configData
     */
    public Calculator( ConfigData configData )
    {
        this.configData = configData;
    }

    /**
     * Convert speed and duration into possible planet positions
     * @param planet
     * @param speed
     * @param unit
     * @param duration
     * @return
     */
    public Collection<SaveTarget> getSaveTargets( Planet origin, FleetSpeed speed, String unit, long duration, Accountable account )
    {
        long maxSpeed = speed.getVelocity();
        Set<SaveTarget> targets = new HashSet<SaveTarget>();
        Position originPosition = origin.getPosition();
        double speedFactor = (account != null) ? account.getFleetSpeed() : 1.0;
        // for spectral mode we need to fiddle with the linear modifiers (pt. 1)
        // unfortunately, the obvious solution, adapting speedFactor, introduces rounding errors
        int spectral = configData.isSpectralMode() ? 10 : 1;
        // from maxSpeed and duration we get a set of 10 distances which we convert into planet positions
        // for additional results  we may want to vary the time frame a bit
        for( int fudge = -240; fudge < 241; fudge += 10 )
        {
            if( duration + fudge > 0 )
            {
                // get distance for the rated speed (100%), % modifiers are applied linear
                int fullDistance = calculateDistance( maxSpeed, duration + fudge, speedFactor );
                // speed modifier 1->10%, 2->20%, 3->30%, ..., 10->100%
                for( int i = 10; i > 0; i-- )
                {
                    // apply speed modifier
                    int distance = i * fullDistance / 10 / spectral;
                    addSaveTarget(targets, origin, originPosition, distance, maxSpeed, speedFactor, speed, unit, duration, account, i);
                    // try a little extra for low speeds
                    if( i == 1 && distance > Distance.baseBetweenSystems && distance < Distance.baseBetweenSystems + Distance.incrementBetweenSystems )
                    {
                        distance = Distance.baseBetweenSystems + Distance.incrementBetweenSystems + 1;
                        addSaveTarget(targets, origin, originPosition, distance, maxSpeed, speedFactor, speed, unit, duration, account, i);
                    }
                }
            }
        }
        return targets;
    }

    private void addSaveTarget(Set<SaveTarget> targets, Planet origin, Position originPosition, int distance, long maxSpeed, double speedFactor, FleetSpeed speed, String unit, long duration, Accountable account, int i)
    {
        Version version = configData.getVersion();
        boolean isOutpost = configData.isOutpost();
        // for spectral mode we need to fiddle with the linear modifiers (pt. 2)
        int spectral = configData.isSpectralMode() ? 10 : 1;
        PositionSet localSystem = getLocalSystem(originPosition, version);
        List<Position> positions = Distance.getNearestPositions( version, originPosition, localSystem, distance );
        for( Position position : positions )
        {
            Planet planet = configData.getPlanetOnPosition( position );
            // calculate actual duration
            for(long actualDistance : Distance.getDistance( version, originPosition, position ))
            {
                // apply speed modifier / revert spectral adaptation
                long actualDuration = spectral * 10 * calculateDuration( maxSpeed, actualDistance, speedFactor ) / i;
                long diff = Math.abs(actualDuration - duration);
                // do not consider targets which are more than 2 hours or more than 100% off
                if( diff < interval && diff < duration )
                {
                    SaveTarget target = new SaveTarget( origin, planet, isLongDistance(origin, planet, distance), speed, 10 * i, unit, actualDuration, duration, account );
                    // restrict outpost travel
                    if(!isOutpost || origin.getPosition().getGalaxy() == version.getOutpostGalaxy() && planet.getPosition().getGalaxy() == version.getOutpostGalaxy()) {
                        targets.add( target );
                    }
                }
            }
        }
    }

    protected static boolean isLongDistance( Planet origin, Planet target, long distance ) {
        // 'long' distance applies if it is in the same galaxy and more than halfway across (>10.800)
        return (origin.getPosition().getGalaxy() == target.getPosition().getGalaxy()) && distance > 10_800;
    }

    private PositionSet getLocalSystem( Position originPosition, Version version ) {
        PositionSet localSystem;
        switch(version) {
            case sirius:
            case genesis:
            case infinity:
                localSystem = SystemData.getVisiblePlanets(version, new Position(originPosition.getGalaxy(), originPosition.getSystem(), 0));
                break;
            default:
                localSystem = new PositionSet(originPosition.getGalaxy(), originPosition.getSystem());
                break;
        }
        return localSystem;
    }

    public Collection<? extends SaveTarget> getSaveTargets2( Planet origin, FleetSpeed maxSpeed, String unit, long duration, Accountable account )
    {
        double speedFactor = (account != null) ? account.getFleetSpeed() : 1.0;
        // for spectral mode we need to fiddle with the linear modifiers
        // unfortunately, the obvious solution, adapting speedFactor, introduces rounding errors
        int spectral = configData.isSpectralMode() ? 10 : 1;
        Set<SaveTarget> targets = new HashSet<SaveTarget>();
        Coordinates originCoordinates = configData.getCoordinates( origin.getPosition() );
        if( duration < maxSpeed.getPreptime() )
        {
            duration = maxSpeed.getPreptime();
        }

        // speed modifier is applied evenly to velocity and acceleration
        // speed modifier 1->10%, 2->20%, 3->30%, ..., 10->100%
        for( int i = 10; i > 0; i-- )
        {
            double speed = i * (double)maxSpeed.getVelocity() / 10.0 / spectral;
            double acceleration = i * (double)maxSpeed.getAcceleration() / 10.0 / spectral;
            double adjustedDuration = duration * speedFactor - maxSpeed.getPreptime();
            long distance = calculateDistance2( speed, acceleration, adjustedDuration );
            // at this distance in a circle around origin lie our targets
            if( distance > 0 )
            {
                List<Coordinates> positions = configData.getTargets( originCoordinates, distance );
                for( Coordinates position : positions )
                {
                    Planet planet = configData.getPlanetOnPosition( position.getPosition() );
                    // calculate actual duration
                    long actualDistance = position.distance( originCoordinates );
                    // apply speed modifier
                    long actualDuration = calculateDuration2( speed, acceleration, actualDistance );
                    // adjust back
                    actualDuration = Math.round( ( actualDuration + maxSpeed.getPreptime() ) / speedFactor );
                    long diff = Math.abs(actualDuration - duration);
                    // do not consider targets which are more than 20 min or more than 50% off
                    if( diff < interval / 6 && diff < duration / 2 )
                    {
                        SaveTarget target = new SaveTarget( origin, planet, false, maxSpeed, 10 * i, unit, actualDuration, duration, account );
                        targets.add( target );
                    }
                }
            }
        }
        return targets;
    }

    /**
     * Return duration in seconds, however, only half of it, b/c total time is there and back
     * @param hour
     * @param minute
     * @return
     * @throws NumberFormatException
     */
    public static long parseTime( String day,  String hour, String minute ) throws NumberFormatException
    {
        long rv;
        long d = 0;
        long h = 0;
        long m = 0;

        if( day != null && day.trim().length() > 0 )
        {
            d = Long.parseLong( day );
        }

        if( hour != null && hour.trim().length() > 0 )
        {
            h = Long.parseLong( hour );
        }

        if( minute != null && minute.trim().length() > 0 )
        {
            m = Long.parseLong( minute );
        }

        rv = ( secondsPerDay * d + secondsPerHour * h + secondsPerMinute * m );
        return rv;
    }

    /**
     * Formats the seconds as time string
     * This is the companion to parseTime() and therefore doubles the amount before formatting
     * @param duration
     * @return
     * @throws NumberFormatException
     */
    public String formatTime( double time )
    {
        StringBuilder sb = new StringBuilder();
        // get back the time 'lost' in parseTime()
        long duration = (long)(2 * time);

        if( duration >= secondsPerDay )
        {
            sb.append( duration / secondsPerDay );
            sb.append( configData.getI18nText( "day.symbol" ) );
            sb.append(  ' ' );
            duration %= secondsPerDay;
        }
        if( duration >= secondsPerHour )
        {
            long h = duration / secondsPerHour;
            if( h < 10 )
            {
                sb.append( '0' );
            }
            sb.append( h );
            duration %= secondsPerHour;
        }
        else
        {
            sb.append( "00" );
        }
        sb.append( ':' );
        if( duration >= secondsPerMinute )
        {
            long m = duration / secondsPerMinute;
            if( m < 10 )
            {
                sb.append( '0' );
            }
            sb.append( m );
            duration %= secondsPerMinute;
        }
        else
        {
            sb.append( "00" );
        }
        sb.append( ':' );
        if( duration < 10 )
        {
            sb.append( '0' );
        }
        sb.append( duration );

        return sb.toString();
    }

    public long getTeleporterPartikel(Position source, Position target)
    {
        long particle = 0;
        if( configData.getVersion() != Version.v20 )
        {
            particle = particle1(source, target);
        }
        else
        {
            particle = particle2(source, target);
        }
        return particle;
    }


    /**
	 * 300
     *--------------- * Entfernung = Flugzeit in sek.
     * wurzel (speed)
     *
     * However, this is only valid for the rated speed (100%), modifiers are applied linear
     *
	 * @param speed
	 * @param distance
	 * @return
	 */
    protected static long calculateDuration( double speed, long distance, double speedFactor )
    {
        // too big by 100, fine tune delta
        long intermediate = (long)Math.ceil(30000.0 * distance / Math.sqrt( speed ) / speedFactor);
        long rv = (long)Math.ceil( intermediate / 100.0 );
        return rv;
    }

    /**
     * Returns the pure flight time to target in seconds, Preparation time must be added on top.
     *
     * @param speed         max. speed, already adjusted by Tech% and speed%
     * @param acceleration  max. acceleration, already adjusted by Tech%
     * @param distance      adjusted distance to target
     * @return
     */
    protected static long calculateDuration2( double speed, double acceleration, double distance )
    {
        long rv;
        double acceleration_time;
        double acceleration_distance = speed * speed / acceleration;   // distance to full speed and back to 0
        if( acceleration_distance >= distance )
        {
            acceleration_time = Math.sqrt( distance / acceleration );
            rv = (long)( 2 * acceleration_time + 1.0 );
        }
        else
        {
            acceleration_time = 2 * speed / acceleration;
            double coasting_distance = distance - acceleration_distance;
            double coasting_time =  coasting_distance / speed;
            rv = (long)( acceleration_time + coasting_time + 1.0 );
        }
        return rv;
    }

    /**
     *
     * However, this is only valid for the rated speed (100%), modifiers are applied linear
     *
     * @param speed
     * @param duration
     * @return
     */
    protected static int calculateDistance( double speed, long duration, double speedFactor )
    {
         long rv = Math.round( speedFactor * duration * Math.sqrt( speed ) / 300.0 );
         return (int)rv;
    }

    /**
     *
     * @param speed         max. speed, already adjusted by Tech% and speed%
     * @param acceleration  max. acceleration, already adjusted by Tech%
     * @param duration      adjusted flight time
     * @return
     */
    protected static long calculateDistance2( double speed, double acceleration, double duration )
    {
        long rv;
        double acceleration_distance;
        double acceleration_time = speed / acceleration;    // time to full speed
        if( 2 * acceleration_time >= duration )
        {
            acceleration_distance = acceleration * duration * duration / 4.0;
            rv = Math.round( acceleration_distance );
        }
        else
        {
            acceleration_distance = acceleration * acceleration_time * acceleration_time;
            double coasting_time =  duration - 2 * acceleration_time;
            double coasting_distance = speed * coasting_time;
            rv = Math.round( acceleration_distance + coasting_distance );
        }
        return rv;
    }

    /**
     *
     * @param distance
     * @param duration
     * @return
     */
    protected static double calculateSpeed( long distance, long duration, double speedFactor )
    {
        double rv = Math.pow( (300.0 * distance / duration / speedFactor ), 2.0 );
        return rv;
    }

    private int particle1( Position source, Position target )
    {
        int distance = Math.max( Distance.galaxyDistance( configData.getVersion(), source, target ), 1 );
        return 50000 * distance;
    }

    private long particle2( Position source, Position target )
    {
        Coordinates one = configData.getCoordinates( source );
        Coordinates two = configData.getCoordinates( target );
        double distance = one.unitDistance( two );
        long particle = (long)Math.ceil( distance * 10 )+ 27500;
        return ( particle < 50000 ) ? 50000 : particle;

    }

//    public static void main(String[] args)
//    {
//        long dist[] = {
//                        11322406988l,
//                        15645280439l,
//                        21908633572l,
//                        75688763367l,
//                        80997636077l,
//                        96136033307l,
//                        124162803307l,
//                        133415847072l,
//                        147180006879l
//        };
//        // expected results: http://si2forum.spaceinvasion.de/index.php?page=Thread&postID=2980#post2980
//        for(long d : dist)
//        {
//            double distance = (double)d / (double)Coordinates.scale;
//            long particle = (long)Math.ceil( distance * 10 )+ 27500;
//            particle = (particle < 50000) ? 50000 : particle;
//            System.out.println(d + "\t\t" + particle);
//        }
//    }

}