/****************************************
 *  COPYRIGHT (C) 2008-2021
 *  Holger Graf
 ****************************************/
package sitool.savecalc;

import java.util.EnumSet;

import sitool.components.Accountable;
import sitool.components.FleetSpeed;
import sitool.components.Planet;
import sitool.components.ProductionBonus;
import sitool.components.Research;
import sitool.components.Specialization;
import sitool.components.Technology;
import sitool.components.Version;


/**
 * @author graf
 *
 */
public class SaveTarget
{
    // a save target is a combination of planet, velocity (in % of full and the unit if known) and flight duration
    private Planet origin;
    private Planet planet;
    private boolean longDistance;
    private int velocity;
    private String unit;
    private long duration;
    private long requestedDuration;
    private FleetSpeed speed;
    private Accountable account;

    public static class AccountLight implements Accountable
    {
        private final String id;
        private final Version version;
        private double fleetSpeed = 1.0;
        private Technology tech = new Technology();
        private Specialization specialization = new Specialization();
        private ProductionBonus productionBonus = new ProductionBonus();

        private AccountLight( Accountable that )
        {
            if( that != null )
            {
                this.id = that.getId();
                this.fleetSpeed = that.getFleetSpeed();
                this.version = that.getVersion();
                this.specialization = new Specialization(that.getSpecialization());
                this.productionBonus = new ProductionBonus(that.getProductionBonus());
                copyTech(that.getMyTech());
            }
            else {
                this.id = "";
                this.version = Version.retro;
            }
        }

        private void copyTech( Technology that )
        {
            for( Research research : EnumSet.of( Research.rocket, Research.diffusion, Research.warp ) )
            {
                tech.setTech( research, that.getTech( research ) );
            }
        }

        /**
         * @param id
         */
        public AccountLight( String id, Version version, Technology tech, Specialization specialization )
        {
            this.id = id;
            this.version = version;
            this.tech = tech;
            this.specialization = specialization;
        }

        @Override
        public String getId()
        {
            return id;
        }

        @Override
        public double getFleetSpeed()
        {
            return fleetSpeed;
        }

        @Override
        public Technology getMyTech()
        {
            return tech;
        }

        @Override
        public Version getVersion()
        {
            return version;
        }

        @Override
        public Specialization getSpecialization() {
            return specialization;
        }

        @Override
        public ProductionBonus getProductionBonus() {
            return productionBonus;
        }
    }

    /**
     * @param planet
     * @param velocity
     * @param duration
     */
    public SaveTarget( Planet origin, Planet planet, boolean longDistance, FleetSpeed speed, int velocity, String unit, long duration, long requestedDuration, Accountable account )
    {
        this.origin = origin;
        this.planet = planet;
        this.longDistance = longDistance;
        this.speed = speed;
        this.velocity = velocity;
        this.unit = unit;
        this.duration = duration;
        this.requestedDuration = requestedDuration;
        this.account = new AccountLight( account );
    }

    /**
     * @return the account
     */
    public Accountable getAccount()
    {
        return account;
    }

    /**
     * @return the origin
     */
    public Planet getOrigin()
    {
        return origin;
    }

    /**
     * @return the planet
     */
    public Planet getPlanet()
    {
        return planet;
    }

    public boolean isLongDistance() {
        return longDistance;
    }
    /**
     * @return the fleet speed
     */
    public FleetSpeed getSpeed()
    {
        return speed;
    }

    /**
     * @return the velocity
     */
    public int getVelocity()
    {
        return velocity;
    }

    /**
     * @return the unit
     */
    public String getUnit()
    {
        return unit;
    }

    /**
     * @return the duration
     */
    public long getDuration()
    {
        return duration;
    }

    /**
     * @return the duration delta
     */
    public long getDelta()
    {
        return Math.abs( duration - requestedDuration );
    }

    public long getRequestedDuration()
	{
		return requestedDuration;
	}

	/**
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals( Object obj )
    {
        boolean rv = false;

        if( obj instanceof SaveTarget )
        {
            SaveTarget that = (SaveTarget)obj;
            rv = origin.equals( that.origin );
            rv &= planet.equals( that.planet );
            rv &= ( duration == that.duration );
        }
        return rv;
    }


    /**
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        int rv = planet.hashCode() + (int)duration;
        return rv;
    }

    /**
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        sb.append( origin.toString() );
        sb.append( ' ' );
        sb.append( planet.toString() );
        sb.append( ' ' );
        sb.append( velocity );
        sb.append( "% " );
        sb.append( unit );
        sb.append( ' ' );
        sb.append( duration );
        return sb.toString();
    }

}
