/****************************************
 *  COPYRIGHT (C) 2006-2021
 *  Holger Graf
 ****************************************/
package sitool.table;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.Icon;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;

import sitool.ConfigData;
import sitool.MainFrame;
import sitool.components.ListItem;
import sitool.components.Building;
import sitool.components.Cost;
import sitool.components.Planet;
import sitool.components.Resource;
import sitool.components.Ship;
import sitool.components.Stage;
import sitool.components.Version;
import sitool.resources.ResourceCalculator;
import sitool.technology.TechtreeDependency;

/**
 * @author graf
 *
 */
public class BuildListTableModel extends StatisticsTableModel implements MouseListener
{
    private static final long serialVersionUID = -3137885733068377331L;
    private List<ListItem<Building>> data = new ArrayList<ListItem<Building>>();

    public final static int TIME = 7;
    public final static int TRANSMITTER = 8;
    public final static int POINTS = 9;
    public final static int DOWN = 10;
    public final static int UP = DOWN + 1;

    private TechtreeDependency techtreeDependency;
    private static Icon upArrow;
    private static Icon downArrow;

    private Map<Long, HashMap<BuildingID, Double>> timeTable = new HashMap<Long, HashMap<BuildingID, Double>>();
    private boolean isConstructorEnabled = false;
    private boolean passStarted = false;
    private Long markedKey = null;

    public BuildListTableModel( MainFrame mainFrame )
    {
        super( mainFrame );
        upArrow = ConfigData.getIcon( "/button_up.gif" );
        downArrow = ConfigData.getIcon( "/button_down.gif" );
        techtreeDependency = new TechtreeDependency( configData );
    }

    public int getColumnCount()
    {
        return UP + 1;
    }

    public int getRowCount()
    {
        return data.size() + 2;
    }

    public String getColumnName( int columnIndex )
    {
        String rv ="";
        switch( columnIndex )
        {
            case 0:
                rv = formatHeader( configData.getI18nText( "Planet" ) );
                break;
            case 1:
                rv = formatHeader( configData.getI18nText( "Building" ) );
                break;
            case 2:
                rv = formatHeader( configData.getI18nText( "Level" ) );
                break;
            case 3:
            case 4:
            case 5:
            case 6:
                rv = formatHeader( configData.getI18nText( Resource.values()[columnIndex - 3].name() ) );
                break;
            case TIME:
                rv = formatHeader( configData.getI18nText( "Buildtime" ).split( " ", 2 ) );
                break;
            case TRANSMITTER:
                rv =  formatHeader( configData.getI18nText( "Transmitters" ).split( " ", 2 ) );
                break;
            case POINTS:
                rv =  formatHeader( configData.getI18nText( "Points" ).split( " ", 2 ) );
                break;
        }
        return rv;
    }

    public Object getValueAt( int rowIndex, int columnIndex )
    {
        Object rv = "";
        if( rowIndex < data.size() )
        {
            ListItem<Building> item = data.get( rowIndex );
            Cost cost = ResourceCalculator.calculateCost( item.getItem(), item.getLevel() );
            switch( columnIndex )
            {
                case 0:
                    rv = item.getPlanet();
                    break;
                case 1:
                    rv = configData.getI18nText( item.getItem().name() );
                    break;
                case 2:
                    rv = Integer.valueOf( item.getLevel() );
                    break;
                case 3:
                case 4:
                case 5:
                case 6:
                    rv = format( cost.getResource( Resource.values()[columnIndex - 3] ) );
                    break;
                case TIME:
                    Version version = configData.getVersion();
                    rv = formatTime( rowIndex, ResourceCalculator.buildTime( cost, isConstructorEnabled, version ) / configData.getSpeedFactor(Building.class) );
                    break;
                case TRANSMITTER:
                    long l = cost.getIron() + cost.getMetal() + cost.getKryptonite() + cost.getSpice();
                    if( l > 0 )
                    {
                        int transmittercapacity = (100 + configData.getSpecialization().getCapacitybonus()) * Ship.transmitter.getLoad() / 100;
                        l /= transmittercapacity;
                        l++;
                    }
                    rv = format( l );
                    break;
                case POINTS:
                    rv = format( (long)cost.points() );
                    break;
                case DOWN:
                    if( rowIndex < data.size() - 1 )
                    {
                        rv = downArrow;
                    }
                    break;
                case UP:
                    if( rowIndex > 0 )
                    {
                        rv = upArrow;
                    }
                    break;
             }
        }
        else
        {
            if( rowIndex == data.size() )
            {
                switch( columnIndex )
                {
                    case 3:
                    case 4:
                    case 5:
                    case 6:
                    case TRANSMITTER:
                    case POINTS:
                        Cost cost = new Cost( 0, 0, 0, 0 );
                        for( int index = 0; index < data.size(); index++ )
                        {
                            ListItem<Building> item = data.get( index );
                            cost.add( ResourceCalculator.calculateCost( item.getItem(), item.getLevel() ) );
                        }
                        if( columnIndex <= 6 )
                        {
                            rv = format( cost.getResource( Resource.values()[columnIndex - 3] ) );
                        }
                        else if( columnIndex == TRANSMITTER )
                        {
                            long l = cost.getIron() + cost.getMetal() + cost.getKryptonite() + cost.getSpice();
                            if( l > 0 )
                            {
                                int transmittercapacity = (100 + configData.getSpecialization().getCapacitybonus()) * Ship.transmitter.getLoad() / 100;
                                l /= transmittercapacity;
                                l++;
                            }
                            rv = format( l );
                        }
                        else  if( columnIndex == POINTS )
                        {
                            rv = format( (long)cost.points() );
                        }
                        break;
                    case TIME:
                        if( passStarted )
                        {
                            // paint result
                            double max = 0;
                            for( Double value : timeTable.get( markedKey ).values() )
                            {
                                max += value.doubleValue();
                            }
                            rv = formatTime( max );
                            // last in row, finish
                            passStarted = false;
                        }
                        else
                        {
                            // since we arrived here the individual build times are already in timeTable
                            markedKey = null;
                            double max = 0;
                            for( Map.Entry<Long, HashMap<BuildingID, Double>> entry : timeTable.entrySet() )
                            {
                                double total = 0;
                                for( Double value : entry.getValue().values() )
                                {
                                    total += value.doubleValue();
                                }
                                if( total > max )
                                {
                                    max = total;
                                    markedKey = entry.getKey();
                                }
                            }
                            if( markedKey != null )
                            {
                                passStarted = true;
                                rv = formatTime( max );
                                // trigger column update to mark the entries that make up max
                                TableModelEvent e = new TableModelEvent( this, 0, getRowCount() - 1, columnIndex );
                                fireTableChanged( e );
                            }
                        }
                        break;
                }
            }
            else
            {
                switch( columnIndex )
                {
                    case 0:
                        rv = configData.getI18nText( "resourceequivalenz" );
                        break;
                    case 3:
                    case 4:
                    case 5:
                    case 6:
                        Cost cost = new Cost( 0, 0, 0, 0 );
                        for( int index = 0; index < data.size(); index++ )
                        {
                            ListItem<Building> item = data.get( index );
                            cost.add( ResourceCalculator.calculateCost( item.getItem(), item.getLevel() ) );
                        }
                        double value = getWeightedResourceEquivalenz( cost );
                        int weight = Resource.values()[columnIndex-3].getWeight();
                        rv = format( Math.round( weight * value ) );
                        break;
                    case TRANSMITTER:
                    case POINTS:
                    case TIME:
                        break;
                }
            }
        }
        return rv;
    }

    public void addItem( ListItem<Building> item )
    {
        data.add( item );
        fireTableDataChanged();
    }

    public void removeAll()
    {
        data.clear();
        passStarted = false;
        fireTableDataChanged();
    }

    public void removeRow( int row )
    {
        data.remove( row );
        passStarted = false;
        fireTableDataChanged();
    }

    public boolean markResource( int row, int column )
    {
        boolean rv = false;
        if( column == TIME && row < data.size() && markedKey != null )
        {
            if( timeTable.size() > 1 )
            {
                // mark only if more than 1 planet is present
                ListItem<Building> item = data.get( row );
                rv = markedKey.equals( item.getKey() );
            }
        }
        return rv;
    }

    public void mouseClicked( MouseEvent e )
    {
        Object o = e.getSource();
        if( o instanceof JTable )
        {
            JTable table = (JTable)o;
            int row = table.rowAtPoint(e.getPoint());
            int col = table.columnAtPoint(e.getPoint());
            // convert to model
            col = table.convertColumnIndexToModel( col );
            if ( SwingUtilities.isLeftMouseButton(e) && e.getClickCount() == 1 )
            {
                if( col == UP || col == DOWN )
                {
                    int start = -1;
                    int stop = -1;
                    switch( col )
                    {
                        case UP:
                            start = row - 1;
                            stop = row + 1;
                            break;
                        case DOWN:
                            start = row;
                            stop = row + 2;
                            break;
                    }

                    if( start >= 0 && stop <= data.size() )
                    {
                        if( enableSwap( start, stop - 1 ) )
                        {
                            // switch them over
                            Collections.rotate( data.subList( start, stop ), 1 );
                            if( col == UP)
                            {
                                table.getSelectionModel().setSelectionInterval( start, start );
                            }
                            else
                            {
                                table.getSelectionModel().setSelectionInterval( stop - 1, stop - 1 );
                            }
                            fireTableDataChanged();
                        }
                    }
                }
            }
        }
    }

    @Override
    public void fireTableDataChanged()
    {
        // at the start of each table update clear the timeTable
        timeTable.clear();
        super.fireTableDataChanged();
    }

    public List<ListItem<Building>> getData()
    {
        return data;
    }

    public void setData( List<ListItem<Building>> data )
    {
        this.data.clear();
        if( data != null )
        {
            this.data.addAll( data );
        }
        fireTableDataChanged();
    }

    public void setConstructor( boolean isConstructorEnabled )
    {
        this.isConstructorEnabled = isConstructorEnabled;
        fireTableDataChanged();
    }

    public void mouseEntered( MouseEvent e )
    {}

    public void mouseExited( MouseEvent e )
    {}

    public void mousePressed( MouseEvent e )
    {}

    public void mouseReleased( MouseEvent e )
    {}

    public Planet copyPlanetAndUpdateFromModel( Planet planet, boolean isAstro, List<ListItem<Building>> buildListItems )
    {
        Planet origin = new Planet( planet.getName(), planet.getPosition() );
        Map<Building, Stage> oldMap = planet.getBuildings( isAstro );
        Map<Building, Stage> newMap = origin.getBuildings( false );
        // copy old map to new
        for( Map.Entry<Building, Stage> entry : oldMap.entrySet() )
        {
            Stage value = new Stage(entry.getValue());
            newMap.put( entry.getKey(), value );
            // special case for galaxy scanner
            if( entry.getKey() == Building.developmentCenter )
            {
                origin.addBuilding( entry.getKey(), value, true );
            }
        }
        // scan table and update map
        for( ListItem<Building> item : buildListItems )
        {
            if( item.getPlanet().equals( origin )  )
            {
                Stage value = newMap.get( item.getItem() );
                if( value == null )
                {
                    value = new Stage(item.getItem(), item.getLevel());
                    newMap.put( item.getItem(), value );
                }
                else if( item.getLevel() > value.getLevel() )
                {
                    value.setLevel( item.getLevel() );
                }
                // special case for galaxy scanner
                if( item.getItem() == Building.developmentCenter )
                {
                    origin.addBuilding( item.getItem(), value, true );
                }
            }
        }
        return origin;
    }

    private boolean enableSwap( int pos1, int pos2 )
    {
        boolean rv = true;
        ListItem<Building> i1 = data.get( pos1 );
        ListItem<Building> i2 = data.get( pos2 );
        if( i1.getPlanet().equals( i2.getPlanet() ) && i1.isAstro() == i2.isAstro() )
        {
            // don't allow higher values of same building before lower
            if( i1.getItem() == i2.getItem() && i1.getLevel() < i2.getLevel() )
            {
                rv = false;
            }
            else
            {
                // check tech dependency up to pos1 - 1
                List<Planet> planets = new ArrayList<Planet>();
                Planet planet = configData.getPlanet( i1.getPlanet() );
                List<ListItem<Building>> buildListItems = data.subList( 0, pos1 );
                Planet origin = copyPlanetAndUpdateFromModel( planet, i1.isAstro(), buildListItems );
                planets.add( origin );
                techtreeDependency.setPlanets( planets );
                rv = techtreeDependency.isEnabled( i2.getItem() );
            }
        }
        return rv;
    }

    private String formatTime( int rowIndex, double buildTime )
    {
        // this is the magic part, we need to consider DevCenter and MSA up to here
        ListItem<Building> item = data.get( rowIndex );
        Long key = item.getKey();
        BuildingID buildingID = new BuildingID( item.getItem(), item.getLevel() );
        // special case super computer
        if( item.getItem() == Building.superComputer)
        {
            double speedFactor = configData.getSpeedFactor(Building.class);
            double constructor = (isConstructorEnabled) ? 0.9 : 1.0;
            buildTime = constructor * 24.0 * 60.0 * 60.0 * 1000.0 / speedFactor;
        }
        else
        {
            Planet planet = configData.getPlanet( item.getPlanet() );
            Stage stage;
            stage = planet.getStage( Building.developmentCenter, item.isAstro() );
            int devCenterLevel = ( stage != null ) ? stage.getLevel() : 0;
            stage = planet.getStage( Building.msa, item.isAstro() );
            int msaLevel = ( stage != null ) ? stage.getLevel() : 0;
            for( int i = 0; i < rowIndex; i++ )
            {
                item = data.get( i );
                Planet p = configData.getPlanet( item.getPlanet() );
                if( planet.equals( p ) )
                {
                    if( item.getItem() == Building.developmentCenter && item.getLevel() > devCenterLevel )
                    {
                        devCenterLevel = item.getLevel();
                    }
                    if( item.getItem() == Building.msa && item.getLevel() > msaLevel )
                    {
                        msaLevel = item.getLevel();
                    }
                }
            }
            int activeDevCenterLevel = devCenterLevel + 1;
            // DevC in V2.0 is reduced by 1
            if( configData.getVersion() == Version.v20 && activeDevCenterLevel > 1 )
            {
                activeDevCenterLevel--;
            }
            buildTime /= ( activeDevCenterLevel );
            buildTime = Math.floor( buildTime );
            if(devCenterLevel >= 12) {
                buildTime /= Math.pow( 2, msaLevel );
                buildTime = Math.floor( buildTime );
            }
        }

        // save the time
        HashMap<BuildingID, Double> list = timeTable.get( key );
        if( list == null )
        {
            list = new HashMap<BuildingID, Double>();
            timeTable.put( key, list );
        }
        list.put( buildingID, Double.valueOf( buildTime ) );

        return formatTime( buildTime );
    }

    private class BuildingID
    {
        private int id;

        public BuildingID( Building building, int level )
        {
            id = ( building.ordinal() << 16 ) + level;
        }

        @Override
        public boolean equals(Object obj)
        {
            boolean rv = false;
            if( obj instanceof BuildingID )
            {
                rv = ( id == ((BuildingID)obj).id );
            }
            return rv;
        }

        @Override
        public int hashCode()
        {
            return id;
        }

        @Override
        public String toString()
        {
            return Integer.toString( id );
        }

    }
}
