/****************************************
 *  COPYRIGHT (C) 2008-2017
 *  Holger Graf
 ****************************************/
package sitool.table;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.List;
import java.util.Map;

import javax.swing.JDialog;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;

import sitool.MainFrame;
import sitool.components.Building;
import sitool.components.Cost;
import sitool.components.Planet;
import sitool.components.Stage;
import sitool.components.Version;
import sitool.resources.BuildingBox;
import sitool.resources.ResourceCalculator;

/**
 * @author graf
 *
 */
public class BuildingTableModel extends PlanetHeaderTableModel implements CellEditorListener, MouseListener
{
    private static final long serialVersionUID = -8208551769084656086L;
    public static final int planetLine = 0;
    private int astroLine;
    private int rows;

    public BuildingTableModel( MainFrame mainFrame )
    {
        super( mainFrame, false );
        update();
    }

    /**
     * @see javax.swing.table.TableModel#getColumnCount()
     */
    public int getColumnCount()
    {
        int columns = mainFrame.getConfigData().getMyPlanets().size() + 1;
        return columns;
    }


    /**
     * @see javax.swing.table.AbstractTableModel#getColumnName(int)
     */
    @Override
    public String getColumnName( int columnIndex )
    {
        String rv;
        if( columnIndex == 0 )
        {
            rv = "";
        }
        else
        {
            List<Planet> data = configData.getMyPlanets();
            Planet target = data.get( columnIndex - 1);
            //return a 2 line header, this is cheating, but, hey it works
            return formatHeader( target.getName(), target.getPosition().toString() );
        }
        return rv;
    }


    /**
     * @see javax.swing.table.AbstractTableModel#getColumnClass(int)
     */
    @Override
    public Class<?> getColumnClass( int columnIndex )
    {
        if( columnIndex == 0 )
            return String.class;
        else
            return Number.class;
    }

    /**
     * @see javax.swing.table.TableModel#getRowCount()
     */
    public int getRowCount()
    {
        // provide summary line
        return rows + 1;
    }

    /**
     * @see javax.swing.table.TableModel#getValueAt(int, int)
     */
    public Object getValueAt( int rowIndex, int columnIndex )
    {
        List<Planet> data = configData.getMyPlanets();
        Version version = configData.getVersion();
        Object rv = null;

        if( columnIndex == 0 )
        {
            if( rowIndex == planetLine )
            {
                rv = configData.getI18nText( "Planet" );
            }
            else if( rowIndex < astroLine )
            {
                Building building = Building.planetBuildings(version)[rowIndex - 1];
                rv = configData.getI18nText( building.name() );
            }
            else if( rowIndex == astroLine )
            {
                rv = configData.getI18nText( "Asteroid" );
            }
            else if( rowIndex == rows )
            {
                rv = getAllPoints();
            }
            else
            {
                int i = rowIndex - astroLine - 1;
                Building building = Building.astroBuildings()[i];
                rv = configData.getI18nText( building.name() );
            }
        }
        else
        {
            Planet target = data.get( columnIndex - 1);
            if( rowIndex == planetLine )
            {
                rv = target.getUsedFields() +  "/" + target.getSize();
            }
            else if( rowIndex < astroLine )
            {
                Building building = Building.planetBuildings(version)[rowIndex - 1];
                Stage stage = target.getStage( building, false );
                rv = Integer.valueOf( (stage != null) ? stage.getLevel() : 0 );
            }
            else if( rowIndex == astroLine )
            {
                if( target.getAstroSize() > 0 )
                {
                    rv = target.getUsedAstroFields() +  "/" + target.getAstroSize();
                }
                else
                {
                    rv = "";
                }
            }
            else if( rowIndex == rows )
            {
                rv = getPoints( columnIndex );
            }
            else
            {
                int i = rowIndex - astroLine - 1;
                Building building = Building.astroBuildings()[i];
                Stage stage = target.getStage( building, true );
                rv = Integer.valueOf( (stage != null) ? stage.getLevel() : 0 );
            }
        }
        return rv;
    }

    @Override
    public boolean isCellEditable(int rowIndex, int columnIndex)
    {
        boolean rv;
        rv = ( columnIndex > 0 && rowIndex > planetLine && rowIndex < astroLine );
        if( rowIndex > astroLine && rowIndex < rows )
        {
            List<Planet> data = configData.getMyPlanets();
            Planet target = data.get( columnIndex - 1 );
            rv = target.getAstroSize() > 0;
        }
        return rv;
    }


    @Override
    public void setValueAt(Object value, int rowIndex, int columnIndex)
    {
        List<Planet> data = configData.getMyPlanets();
        Planet target = data.get( columnIndex - 1 );

        Building building = null;
        Stage stage = null;
        boolean astro;
        if( rowIndex < astroLine )
        {
            astro = false;
            building = Building.planetBuildings(configData.getVersion())[rowIndex - 1];
        }
        else
        {
            astro = true;
            int i = rowIndex - astroLine - 1;
            building = Building.astroBuildings()[i];
        }

        int level = Integer.parseInt( value.toString() );

        if( level == 0 )
        {
            target.removeBuilding( building, astro );
        }
        else
        {
            stage = target.getStage( building, astro );
            if( stage == null )
            {
                stage = new Stage(building, level);
                target.addBuilding( building, stage, astro );
            }
            stage.setLevel(level);
        }
    }

    public void update() {
        astroLine = Building.planetBuildings(configData.getVersion()).length + 1;
        rows = Building.planetBuildings(configData.getVersion()).length + Building.astroBuildings().length + 2;
    }

    public int getAstroLine() {
        return astroLine;
    }

    public int getRows() {
        return rows;
    }

    public void editingCanceled( ChangeEvent e )
    {
    }

    public void editingStopped( ChangeEvent e )
    {
        // update the status rows
        fireTableRowsUpdated( planetLine, planetLine );
        fireTableRowsUpdated( astroLine, astroLine );
        fireTableRowsUpdated( rows, rows );
    }

    protected Cost totalCost( Planet planet )
    {
        Cost total = new Cost( 0, 0, 0, 0 );

        // planetary buildings
        for( Map.Entry<Building, Stage> entry : planet.getBuildings( false ).entrySet() )
        {
            Building building = entry.getKey();
            int level = entry.getValue().getLevel();
            total.add( ResourceCalculator.calculateTotalCost( building, level ) );
        }
        // asteroid buildings (Meta-Ridley says Astro Buildings don't count)
        // but Ferox says, they do, now
        for( Map.Entry<Building, Stage> entry : planet.getBuildings( true ).entrySet() )
        {
            Building building = entry.getKey();
            // 19.01.2012 scanner does count
//            // galaxy scanner does not count, though
//            if( Building.scanner == building )
//            {
//                continue;
//            }
            int level = entry.getValue().getLevel();
            total.add( ResourceCalculator.calculateTotalCost( building, level ) );
        }

        return total;
    }

    public void mouseClicked( MouseEvent e )
    {
        Object o = e.getSource();
        if( o instanceof JTable )
        {
            JTable table = (JTable)o;
            int row = table.rowAtPoint(e.getPoint());
            int col = table.columnAtPoint(e.getPoint());
            // convert to model
            row = table.convertRowIndexToModel( row );
            col = table.convertColumnIndexToModel( col );
            if ( SwingUtilities.isLeftMouseButton(e) && e.getClickCount() == 1 )
            {
                if( ( row == planetLine || row == astroLine ) && ( col > 0 ) )
                {
                    List<Planet> data = configData.getMyPlanets();
                    Planet planet = data.get( col - 1 );
                    boolean isAstro = ( row == astroLine );
                    if( !isAstro || planet.getAstroSize() > 0 )
                    {
                        JDialog dlg = new BuildingBox( mainFrame, planet, isAstro );
                        mainFrame.showModalBox( dlg );
                    }
                }
            }
        }
    }

    public void mouseEntered( MouseEvent e )
    {}

    public void mouseExited( MouseEvent e )
    {}

    public void mousePressed( MouseEvent e )
    {}

    public void mouseReleased( MouseEvent e )
    {}
}
