/****************************************
 *  COPYRIGHT (C) 2008, 2022
 *  Holger Graf
 ****************************************/
package sitool.table;

import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.table.AbstractTableModel;

import sitool.Account;
import sitool.ConfigData;
import sitool.MainFrame;
import sitool.components.Building;
import sitool.components.Cost;
import sitool.components.Planet;
import sitool.components.Resource;
import sitool.components.Stage;
import sitool.components.TotalPointDisplay;

/**
 * @author graf
 *
 */
public abstract class StatisticsTableModel extends AbstractTableModel implements ResourceMarker, CellEditorListener
{
    private static final long serialVersionUID = -327932871342205186L;
    private static final long researchNetworkBonus1D= 1000l * 60l * 60l * 24l;

    protected MainFrame mainFrame;
    protected ConfigData configData;

    public StatisticsTableModel( MainFrame mainFrame )
    {
        this.mainFrame = mainFrame;
        configData = mainFrame.getConfigData();
    }

    protected double getWeightedResourceEquivalenz( Cost cost )
    {
        double rv = 0;

        rv += (double)cost.getIron() / Resource.iron.getWeight();
        rv += (double)cost.getMetal() / Resource.metal.getWeight();
        rv += (double)cost.getKryptonite() / Resource.kryptonite.getWeight();
        rv += (double)cost.getSpice() / Resource.spice.getWeight();

        return rv;
    }

    protected Resource getFirstResource( Cost cost )
    {
        Resource resource = null;
        for( Resource r : Resource.values() )
        {
            if( cost.getResource( r ) > 0 )
            {
                resource = r;
                break;
            }
        }

        return resource;
    }

    protected Resource getSecondResource( Cost cost )
    {
        Resource resource = getFirstResource( cost );
        if( resource != null )
        {
            resource = resource.next();
        }
        while( resource != null &&  cost.getResource( resource ) == 0 )
        {
            resource = resource.next();
        }

        return resource;
    }

    protected Resource getThirdResource( Cost cost )
    {
        Resource resource = getSecondResource( cost );
        if( resource != null )
        {
            resource = resource.next();
        }
        while( resource != null &&  cost.getResource( resource ) == 0 )
        {
            resource = resource.next();
        }

        return resource;
    }

    protected Resource getFourthResource( Cost cost )
    {
        Resource resource = getThirdResource( cost );
        if( resource != null )
        {
            resource = resource.next();
        }
        while( resource != null &&  cost.getResource( resource ) == 0 )
        {
            resource = resource.next();
        }

        return resource;
    }

    protected String format( long value )
    {
        String rv;
        if( value == 0 )
        {
            rv = "";
        }
        else
        {
            rv = TotalPointDisplay.format( value );
        }
        return rv;
    }

    protected String formatHeader( String... lines )
    {
        StringBuilder sb = new StringBuilder();
        sb.append( "<html>" );
        sb.append( "<p align=\"left\">" );
        for( int i = 0; i < lines.length || i < 2; i++ )
        {
            if( i > 0 )
            {
                sb.append( "<br>" );
            }
            if( i >= lines.length || lines[i].trim().length() == 0 )
            {
                sb.append( "&nbsp;" );
            }
            else
            {
                sb.append( lines[i] );
            }
        }
        sb.append( "</p>" );
        sb.append( "<html>" );
        return sb.toString();
    }

    protected String formatTime( double buildTime )
    {
        return formatTime(buildTime, RoundMillis.ROUND);
    }

    protected String formatTime( double buildTime, RoundMillis roundMillis )
    {
        StringBuilder sb = new StringBuilder();

        if( buildTime >= 0 )
        {
            long millis = 0;
            switch( roundMillis ) {
                case NONE:
                    millis = (long)Math.ceil( buildTime );
                    break;
                case UP:
                    millis = (long)Math.ceil( buildTime / 1000 ) * 1000;
                    break;
                case ROUND:
                    millis = Math.round( buildTime / 1000 ) * 1000;
                    break;
            }

            // 24 * 60 * 60
            long day = millis  / 86_400_000;
            if( day > 0 )
            {
                sb.append( day );
                sb.append( configData.getI18nText( "day.symbol" ) );
                sb.append(  ' ' );
                millis %= 86_400_000;
            }
            // 60 * 60
            long hour = millis / 3_600_000;
            if( hour < 10 )
            {
                sb.append( '0' );
            }
            sb.append( hour );
            sb.append( ':' );
            millis %= 3_600_000;

            // 60
            long minute = millis / 60_000;
            if( minute < 10 )
            {
                sb.append( '0' );
            }
            sb.append( minute );
            sb.append( ':' );
            millis %= 60_000;

            if( millis < 10_000 )
            {
                sb.append( '0' );
            }
            sb.append( millis / 1_000 );
            if( roundMillis == RoundMillis.NONE ) {
                millis %= 1_000;
                sb.append('.');
                if(millis < 100) {
                    sb.append('0');
                }
                if(millis < 10) {
                    sb.append('0');
                }
                sb.append( millis );
            }

        }
        return sb.toString();
    }


    protected long getResearchCap() {
        long researchCap = 0;
        Account current = configData.getActiveAccount();
        if( current != null ) {
            researchCap = current.getVersion().getResearchCap();
        }
        switch(configData.getVersion()) {
            case sirius:
                for( Planet planet : configData.getMyPlanets()) {
                    Stage  stage = planet.getStage( Building.researchLab, false );
                    int labLevel = ( stage != null ) ? stage.getLevel() : 0;
                    researchCap -= (researchNetworkBonus1D * labLevel);
                }
                // lower bound
                researchCap = (researchCap < 30 * researchNetworkBonus1D) ? 30 * researchNetworkBonus1D : researchCap;
                break;
            default:
                break;
        }
        return researchCap;
    }
    public void editingCanceled( ChangeEvent e )
    {}

    public void editingStopped( ChangeEvent e )
    {
        fireTableDataChanged();
    }

    protected static enum RoundMillis {
        NONE, ROUND, UP;
    }
}
