/****************************************
 *  COPYRIGHT (C) 2008
 *  Holger Graf
 ****************************************/
package sitool.table;

import java.util.Arrays;

import sitool.MainFrame;
import sitool.components.Cost;
import sitool.components.Resource;
import sitool.components.Ship;
import sitool.components.TotalPointDisplay;

/**
 * @author graf
 *
 */
public class TradeTableModel extends StatisticsTableModel
{
    private static final long serialVersionUID = -3725902375412789785L;
    private boolean[] locks;
    private int[] percent;
    private boolean calculateByValue;
    private Cost inValues;

    public TradeTableModel( MainFrame mainFrame )
    {
        super( mainFrame );
        locks = new boolean[Resource.values().length];
        Arrays.fill(locks, false );
        percent = new int[Resource.values().length];
        Arrays.fill(percent, 100 );
        configData = mainFrame.getConfigData();
    }

    /**
     * @see javax.swing.table.TableModel#getColumnCount()
     */
    public int getColumnCount()
    {
        return 9;
    }

    /**
     * @see javax.swing.table.AbstractTableModel#getColumnName(int)
     */
    @Override
    public String getColumnName( int columnIndex )
    {
        String rv ="";
        switch( columnIndex )
        {
            case 0:
            case 2:
                rv = formatHeader( "" );
                break;
            case 1:
                rv = formatHeader( "%" );
                break;
            case 3:
                rv = formatHeader( configData.getI18nText( "Value" ) );
                break;
            case 5:
                rv = formatHeader( configData.getI18nText( "min" ) );
                break;
            case 7:
                rv = formatHeader( configData.getI18nText( "max" ) );
                break;
            case 4:
            case 6:
            case 8:
                rv =  formatHeader( configData.getI18nText( "Transmitters" ).split( " ", 2 ) );
        }
        return rv;
    }


    /**
     * @see javax.swing.table.TableModel#getRowCount()
     */
    public int getRowCount()
    {
        return Resource.values().length + 1;
    }

    /**
     * @see javax.swing.table.TableModel#getValueAt(int, int)
     */
    public Object getValueAt( int rowIndex, int columnIndex )
    {
        Object rv = "";
        long value = 0;

        if( rowIndex < Resource.values().length )
        {
            Resource resource = Resource.values()[rowIndex];
            boolean show = ( inValues.getResource( resource ) == 0 );
            switch( columnIndex )
            {
                case 0:
                    rv = configData.getI18nText( Resource.values()[rowIndex].name() );
                    break;
                case 1:
                    int p = 0;
                    if( show )
                    {
                        p = percent[rowIndex];
                    }
                    rv = Integer.valueOf( p );
                    break;
                case 2:
                    if( show && !calculateByValue )
                    {
                        rv = Boolean.valueOf( locks[rowIndex] );
                    }
                    break;
                case 3:
                case 5:
                case 7:
                    if( show )
                    {
                        value = calculate( rowIndex, columnIndex );
                        if( value > 0 )
                        {
                            rv = format( value );
                        }
                    }
                    break;
                case 4:
                case 6:
                case 8:
                    if( show )
                    {
                        value = calculate( rowIndex, columnIndex - 1 );
                        if( value > 0 )
                        {
                            int transmittercapacity = (100 + configData.getSpecialization().getCapacitybonus()) * Ship.transmitter.getLoad() / 100;
                            value /= transmittercapacity;
                            value++;
                            rv = format( value );
                        }
                    }
                    break;
            }
        }
        else
        {
            switch( columnIndex )
            {
                case 3:
                case 5:
                case 7:
                    value = 0;
                    for( int row = 0; row < Resource.values().length; row++)
                    {
                        Resource resource = Resource.values()[row];
                        boolean show = ( inValues.getResource( resource ) == 0 );
                        if( show )
                        {
                            value += calculate( row, columnIndex );
                        }
                    }
                    if( value > 0 )
                    {
                        rv = format( value );
                    }
                    break;
                case 4:
                case 6:
                case 8:
                    value = 0;
                    for( int row = 0; row < Resource.values().length; row++)
                    {
                        Resource resource = Resource.values()[row];
                        boolean show = ( inValues.getResource( resource ) == 0 );
                        if( show )
                        {
                            value += calculate( row, columnIndex - 1 );
                        }
                    }
                    if( value > 0 )
                    {
                        int transmittercapacity = (100 + configData.getSpecialization().getCapacitybonus()) * Ship.transmitter.getLoad() / 100;
                        value /= transmittercapacity;
                        value++;
                        rv = format( value );
                    }
                    break;
            }
        }
        return rv;
    }

    @Override
    public boolean isCellEditable(int rowIndex, int columnIndex)
    {
        boolean rv = false;
        if( rowIndex < Resource.values().length )
        {
            switch( columnIndex )
            {
                case 1:
                    rv = !locks[rowIndex];
                    // however, there is a slight wrinkle
                    // the last unlocked cell may not be edited, too
                    if( rv )
                    {
                        int unlocked = 0;
                        Resource[] resources = Resource.values();
                        for( int i = 0; i < resources.length; i++ )
                        {
                            if( !locks[i] && inValues.getResource(resources[i]) == 0 )
                            {
                                unlocked++;
                            }
                        }
                        rv = ( unlocked > 1 );
                    }
                    break;
                case 2:
                    rv = !calculateByValue;
                    break;
            }
        }
        return rv;
    }


    @Override
    public void setValueAt(Object value, int rowIndex, int columnIndex)
    {
        switch( columnIndex )
        {
            case 0:
                break;
            case 1:
                if( value instanceof Number )
                {
                    percent[rowIndex] = ((Number)value).intValue();
                }
                break;
            case 2:
                if( value != null )
                {
                    locks[rowIndex] = Boolean.valueOf( value.toString() );
                }
                break;
            case 3:
                break;
            case 4:
                break;
            case 5:
                break;
        }
        if( !calculateByValue )
        {
            adjustVolume( rowIndex );
        }

        fireTableDataChanged();
    }

    public void update( boolean isValue, Cost inValues )
    {
        calculateByValue = isValue;
        this.inValues = inValues;
        if( !calculateByValue )
        {
            adjustVolume( -1 );
        }
        else
        {
            Arrays.fill( percent, 100 );
            Arrays.fill( locks, false );
        }
        fireTableDataChanged();
    }

    private long calculate( int rowIndex, int columnIndex )
    {
        long rv = 0;

        if( calculateByValue )
        {
            rv = calculateByValue( rowIndex, columnIndex );
        }
        else
        {
            rv = calculateByQuantity( rowIndex, columnIndex );
        }
        return rv;
    }

    private long calculateByQuantity( int rowIndex, int columnIndex )
    {
        double value = 0;
        int outValues = 0;
        // get input (normalized)
        for( Resource resource : Resource.values() )
        {
            double in = (double)inValues.getResource( resource ) / (double)resource.getWeight();
            if( in == 0 && percent[resource.ordinal()] > 0 )
            {
                outValues++;
            }
            value += in;
        }

        double factor = 0.0;
        switch( outValues )
        {
            case 0:
                factor = 0.0;
                break;
            case 1:
                if( percent[rowIndex] > 0 )
                {
                    percent[rowIndex] = 100;
                    factor = 1.0;
                }
                break;
            case 2:
                factor = factor2( rowIndex );
                break;
            case 3:
                factor = factor3( rowIndex );
                break;
        }

        Resource resource = Resource.values()[rowIndex];
        // apply weight for output
        value *= resource.getWeight() * factor;
        double d;
        switch( columnIndex )
        {
            case 5:
                d = 0.8;
                break;
            case 7:
                d = 1.2;
                break;
            case 3:
            default:
                d = 1.0;
                break;
        }

        // apply percent value
        value *= d;
        return Math.round( value );
    }

    private double factor3( int rowIndex )
    {
        double factor = 0.0;
        // 3 outs means 1 in; reverse
        Resource[] resources = Resource.values();
        for( int i = 0; i < resources.length; i++ )
        {
            long value = inValues.getResource( resources[i] );
            if( value == 0 )
            {
                factor += (double)percent[i] / (double)resources[i].getWeight();
            }
        }
        // now this is what we get from our sample
        // apply to requested value
        factor = percent[rowIndex] / factor;
        // this is already final, therefore remove the weight
        factor /= resources[rowIndex].getWeight();
        return factor;
    }

    private double factor2( int rowIndex )
    {
        double factor = 0.0;
        Resource[] resources = Resource.values();
        if( percent[rowIndex] > 0 )
        {
            // find another value
            int other = 0;
            for( int i = 0; i < resources.length; i++ )
            {
                if( i != rowIndex && percent[i] > 0 && inValues.getResource( resources[i] )  == 0 )
                {
                    // there IS another one b/c percent < 100
                    other = i;
                    break;
                }
            }
            double x = ( percent[rowIndex] * resources[other].getWeight() );
            double y = ( percent[other] * resources[rowIndex].getWeight() );
            x /= y;
            double part = (percent[rowIndex] + percent[other] ) / 100.0;
            factor = x / ( part  + x );
        }
        return factor;
    }

    private long calculateByValue( int rowIndex, int columnIndex )
    {
        double value = 0;
        // get input (normalized)
        for( Resource resource : Resource.values() )
        {
            value += (double)inValues.getResource( resource ) / (double)resource.getWeight();
        }
        Resource resource = Resource.values()[rowIndex];
        // apply weight for output
        value *= resource.getWeight();
        double d;
        switch( columnIndex )
        {
            case 5:
                d = 0.8;
                // keep min. of 80% even if overcharged
                if( percent[rowIndex] > 100 )
                {
                    d /= percent[rowIndex] / 100.0;
                }
                break;
            case 7:
                d = 1.2;
                // keep max. of 120% even if overcharged
                if( percent[rowIndex] > 100 )
                {
                    d /= percent[rowIndex] / 100.0;
                }
                break;
            case 3:
            default:
                d = 1.0;
                break;
        }
        // apply percent value
        value *= d * percent[rowIndex] / 100.0;
        return Math.round( value );
    }

    private void adjustVolume( int current )
    {
        int full = 100;
        int unlocked = 0;
        Resource[] resources = Resource.values();
        boolean[] in = new boolean[resources.length];
        for( int i = 0; i < resources.length; i++ )
        {
            in[i] = ( inValues.getResource( resources[i] ) != 0 || i == current );
            if( ( locks[i] && !in[i] ) || i == current )
            {
                full -= percent[i];
            }
            else if( !in[i] )
            {
                unlocked++;
            }
        }
        if( unlocked > 0 )
        {
            int distribute = full / unlocked;
            int leftover = full % unlocked;
            for( int i = 0; i < locks.length; i++ )
            {
                if( !locks[i] && !in[i] )
                {
                    percent[i] = distribute;
                    if( leftover > 0 )
                    {
                        percent[i]++;
                        leftover--;
                    }
                }
            }
        }
        else if( full > 0 && current >= 0 && current < percent.length )
        {
            percent[current] += full;
        }
    }

    protected String format( long value )
    {
        return TotalPointDisplay.format( value );
    }

    public boolean markResource( int row, int column )
    {
        // not used
        return false;
    }

}
