/****************************************
 *  COPYRIGHT (C) 2021
 *  Holger Graf
 ****************************************/
package sitool.technology;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.border.EtchedBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import sitool.Account;
import sitool.ConfigData;
import sitool.MainFrame;
import sitool.Parser;
import sitool.components.ListItem;
import sitool.components.Building;
import sitool.components.EnumWrapper;
import sitool.components.ModulePane;
import sitool.components.Planet;
import sitool.components.Position;
import sitool.components.Research;
import sitool.components.Stage;
import sitool.resources.ImportExportPane;
import sitool.table.AlternateLineTableCellRenderer;
import sitool.table.BlindTableCellRenderer;
import sitool.table.ImageTableCellRenderer;
import sitool.table.ResourceTableCellRenderer;
import sitool.table.TechBuilderTableModel;
import sitool.table.TooltipTable;

/**
 * @author graf
 *
 */
public class TechBuilderPane extends JPanel implements ModulePane, ImportExportPane {

    private static final long serialVersionUID = 7588823623104936926L;
    protected static final String extension = ".srz";

    private final MainFrame mainFrame;
    private final ConfigData configData;
    private JComboBox<EnumWrapper<Research>> techs = new JComboBox<>();
    private JLabel techLabel = new JLabel();
    private JButton addButton = new JButton();
    private JButton deleteButton = new JButton();
    private JButton deleteAll = new JButton();
    private JButton importButton = new JButton();
    private JButton exportButton = new JButton();
    private JLabel canvas = new JLabel();
    private TechBuilderTableModel model;
    private JTable table;
    private final JPanel center;
    private Account currentAccount;
    private final TechtreeDependency techtreeDependency;

    public TechBuilderPane( MainFrame parent )
    {
        mainFrame = parent;
        configData = mainFrame.getConfigData();
        techtreeDependency = new TechtreeDependency( configData );
        setLayout( new BorderLayout() );
        setBorder( BorderFactory.createEtchedBorder( EtchedBorder.LOWERED ) );

        center = createCenterPanel();
        //center.setPreferredSize( new Dimension( 640, 520 ) );

        add( center, BorderLayout.CENTER );
    }

    @SuppressWarnings("unchecked")
    protected void actionAdd() {
        Research research =  ((EnumWrapper<Research>)techs.getSelectedItem()).getBase();
        int level = configData.getMyTech().getTech(research) + 1;
        Planet planet = getResearchPlanet();
        Planet researchPlanet = new Planet(planet.getName(), planet.getPosition());
        ListItem<Research> item = new ListItem<>( researchPlanet, research, level );
        // adjust level (if necessary)
        List<ListItem<Research>> data = model.getData();
        while( data.contains( item  ) )
        {
            item.increaseLevel();
        }
        model.addItem( item );
        // update techtreeDependency
        techtreeDependency.updateTechnology(model.getData());
    }

    protected void actionDelete()
    {
        int row = table.getSelectedRow();
        model.removeRow( row );
        techtreeDependency.updateTechnology(model.getData());
    }

    @SuppressWarnings("unchecked")
    protected void actionTechChanged( ItemEvent e )
    {
        if( e.getStateChange() == ItemEvent.SELECTED )
        {
            Research research = ((EnumWrapper<Research>)e.getItem()).getBase();

            canvas.setIcon( research.getIcon() );
            canvas.repaint();
            addButton.setEnabled( enableAdd() );
        }
    }

    protected void actionLoad()
    {
        Document doc = configData.readData( this );
        if( doc != null )
        {
            NodeList accounts = doc.getElementsByTagName( Parser.Tag_Account );
            if( accounts.getLength() > 0 )
            {
                Element account = (Element)accounts.item( 0 );
                String accountId = account.getAttribute( Parser.Attribute_Name );
                // load for current account only
                if( currentAccount != null && currentAccount.getId().equals( accountId ) )
                {
                    List<ListItem<Research>> data = new ArrayList<ListItem<Research>>();
                    NodeList buildList = account.getElementsByTagName( Parser.Tag_TechlistItem );
                    if( buildList.getLength() > 0 )
                    {
                        @SuppressWarnings("unchecked")
                        ListItem<Research>[] items = new ListItem[buildList.getLength()];
                        int i;
                        // build build list data
                        for( i = 0; i < items.length; i++ )
                        {
                            Element itemNode = ( Element )buildList.item( i );
                            String name = itemNode.getAttribute( Parser.Attribute_Name );
                            int galaxy = ConfigData.parseIntegerAttribute( itemNode.getAttribute( Parser.Attribute_Galaxy ) );
                            int system = ConfigData.parseIntegerAttribute( itemNode.getAttribute( Parser.Attribute_System ) );
                            int p = ConfigData.parseIntegerAttribute( itemNode.getAttribute( Parser.Attribute_Planet ) );
                            Planet planet = new Planet( name, new Position( galaxy, system, p ) );
                            // look if we still have a planet on this position, the name might have changed
                            Planet myPlanet = configData.getPlanet( planet );
                            if( myPlanet != null )
                            {
                                // name may have changed
                                planet.setName( myPlanet.getName() );
                            }
                            else
                            {
                                planet = null;
                            }
                            if( planet != null )
                            {
                                int level = ConfigData.parseIntegerAttribute( itemNode.getAttribute( Parser.Attribute_Level ) );
                                Research research = ConfigData.parseResearchAttribute( itemNode.getAttribute( Parser.Attribute_Tech ) );

                                int position = ConfigData.parseIntegerAttribute( itemNode.getAttribute( Parser.Attribute_Position ) );
                                items[position] = new ListItem<Research>( planet, research, level );
                            }
                        }
                        for( i = 0; i < items.length; i++ )
                        {
                            // add all non-null items
                            if( items[i] != null )
                            {
                                data.add( items[i] );
                            }
                        }
                    }
                    model.setData( data );
                    getResearchPlanet();
                    techtreeDependency.updateTechnology(model.getData());
                }
            }
        }
    }

    protected void actionSave()
    {
        configData.saveData( this );
    }

    protected JPanel createCenterPanel()
    {
        JPanel center = new JPanel( new BorderLayout() );
        initTechs();
        deleteButton.setEnabled( false );

        JPanel north = new JPanel( new BorderLayout() );
        JPanel northwest = new JPanel( new GridBagLayout() );
        center.add( north, BorderLayout.NORTH );
        north.add( northwest, BorderLayout.WEST );
        north.setBorder( BorderFactory.createEtchedBorder( EtchedBorder.LOWERED ) );

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.gridheight = 1;
        constraints.fill = GridBagConstraints.HORIZONTAL;
        constraints.anchor = GridBagConstraints.NORTH;
        constraints.weightx = 1.0;
        constraints.weighty = 0.0;
        constraints.insets = new Insets( 2, 2, 2, 2 );

        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.gridwidth = 1;
        constraints.weightx = 0.0;
        northwest.add( new JLabel("         "), constraints );
        constraints.gridx = 1;
        constraints.gridy = 0;
        constraints.gridwidth = 3;
        constraints.weightx = 0.0;
        northwest.add( techLabel, constraints );
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.gridwidth = 1;
        constraints.weightx = 0.0;
        northwest.add( new JLabel("         "), constraints );
        constraints.gridx = 1;
        constraints.gridy = 1;
        constraints.gridwidth = 3;
        constraints.weightx = 1.0;
        northwest.add( techs, constraints );
        constraints.gridx = 0;
        constraints.gridy = 2;
        constraints.gridwidth = 3;
        constraints.weightx = 1.0;
        constraints.weighty = 0.0;
        northwest.add( new JLabel( " " ), constraints );
        constraints.gridx = 4;
        constraints.gridy = 3;
        constraints.gridwidth = 3;
        constraints.weightx = 1.0;
        constraints.weighty = 0.0;
        northwest.add( addButton, constraints );

        canvas.setPreferredSize( new Dimension( 120, 120 ) );
        canvas.setHorizontalAlignment( SwingConstants.CENTER );
        north.add( canvas, BorderLayout.CENTER );

        JPanel northeast = new JPanel( new FlowLayout() );
        JPanel buttonPanel = new JPanel( new GridLayout( 5, 1 ) );
        north.add( northeast, BorderLayout.EAST );
        northeast.add( buttonPanel );
        buttonPanel.add( deleteButton );
        buttonPanel.add( deleteAll );
        buttonPanel.add( Box.createVerticalGlue() );
        buttonPanel.add( importButton );
        buttonPanel.add( exportButton );

        model = new TechBuilderTableModel( mainFrame );
        final BlindTableCellRenderer summaryRenderer = new BlindTableCellRenderer();
        summaryRenderer.setHorizontalAlignment( SwingConstants.CENTER );
        final AlternateLineTableCellRenderer leftRenderer = new AlternateLineTableCellRenderer();
        final ResourceTableCellRenderer timeColumnRenderer = new ResourceTableCellRenderer();
        timeColumnRenderer.setHorizontalAlignment( SwingConstants.RIGHT );
        final ImageTableCellRenderer ic = new ImageTableCellRenderer();
        table = new TooltipTable( model )
        {
            private static final long serialVersionUID = 7149497989459592130L;
            public TableCellRenderer getCellRenderer(int row, int column)
            {
                TableCellRenderer rv;
                // convert to model
                int modelColumn = convertColumnIndexToModel( column );
                if( row >= model.getRowCount() - 2 )
                {
                    rv = summaryRenderer;
                }
                else
                {
                    if( modelColumn == 0 )
                    {
                        rv = leftRenderer;
                    }
                    else if( modelColumn == TechBuilderTableModel.TIME )
                    {
                        rv = timeColumnRenderer;
                    }
                    else if( modelColumn == TechBuilderTableModel.DOWN ||
                             modelColumn == TechBuilderTableModel.UP
                           )
                    {
                        rv = ic;
                    }
                    else
                    {
                        rv = super.getCellRenderer( row, column );
                    }
                }
                return rv;
            }
        };
        table.setDefaultRenderer( Object.class, new AlternateLineTableCellRenderer() );
        table.putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);

        table.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
        AlternateLineTableCellRenderer defaultRenderer = new AlternateLineTableCellRenderer();
        defaultRenderer.setHorizontalAlignment( SwingConstants.RIGHT );
        table.setDefaultRenderer( Object.class, defaultRenderer );
        table.addMouseListener( model );

        JScrollPane scrollPane = new JScrollPane( table,
                        JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                        JScrollPane.HORIZONTAL_SCROLLBAR_NEVER );
        center.add( scrollPane, BorderLayout.CENTER );
        scrollPane.setBorder( BorderFactory.createEtchedBorder( EtchedBorder.LOWERED ) );
        center.add( scrollPane, BorderLayout.CENTER );

        table.getSelectionModel().addListSelectionListener( new ListSelectionListener()
        {
            public void valueChanged( ListSelectionEvent e )
            {
                if( !e.getValueIsAdjusting() )
                {
                    boolean mayDelete = false;
                    // single selection, only one index expected
                    int row = table.getSelectedRow();
                    // enable the button when selection is active (but don't delete the summary lines!)
                    mayDelete = ( row >= 0 && row < model.getRowCount() - 2 );
                    // check for dependencies
                    if( mayDelete && row < model.getRowCount() - 3 )
                    {
                        ListItem<Research> item2delete = model.getData().get( row );
                        // higher versions of same building
                        List<ListItem<Research>> data = model.getData().subList( row + 1, model.getRowCount() - 2 );
                        for( ListItem<Research> item : data )
                        {
                            if( item2delete.getPlanet().equals( item.getPlanet() ) )
                            {
                                if( item2delete.getItem() == item.getItem() && item2delete.getLevel() < item.getLevel() )
                                {
                                    mayDelete = false;
                                    break;
                                }
                            }
                        }
                        // tech dependency
                        while( mayDelete && ++row < model.getRowCount() - 3 )
                        {
                            ListItem<Research> item2test = model.getData().get( row );
                            data = new ArrayList<ListItem<Research>>( model.getData().subList( 0, row ) );
                            data.remove( item2delete );
                            techtreeDependency.updateTechnology(data);
                            mayDelete = techtreeDependency.isEnabled( item2test.getItem() );
                        }
                    }
                    deleteButton.setEnabled( mayDelete );
                }
            }
        });
        table.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0), "delete");
        table.getActionMap().put("delete", new AbstractAction() {
            private static final long serialVersionUID = 5999113634471333070L;

            @Override
            public void actionPerformed( ActionEvent e ) {
                if(deleteButton.isEnabled()) {
                    actionDelete();
                }
            }
        });

        techs.addItemListener( new ItemListener() {

            public void itemStateChanged( ItemEvent e )
            {
                actionTechChanged( e );
            }

        });
        addButton.addActionListener( new ActionListener(){
            public void actionPerformed( ActionEvent e )
            {
                actionAdd();
            }
        });
        deleteButton.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent e )
            {
                actionDelete();
            }

        });
        deleteAll.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent e )
            {
                model.removeAll();
                techtreeDependency.updateTechnology(model.getData());
            }
        });
        importButton.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent e )
            {
                actionLoad();
            }
        });
        exportButton.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent e )
            {
                actionSave();
            }
        });
        return center;
    }

    @SuppressWarnings("unchecked")
    private boolean enableAdd()
    {
        boolean rv = false;
        Planet researchPlanet = getResearchPlanet();
        if( researchPlanet != null
         && researchPlanet.getStage(Building.researchLab, false) != null
         && researchPlanet.getStage(Building.researchLab, false).getLevel() > 0
        ) {
            techtreeDependency.updateTechnology(model.getData());
            Research research = ((EnumWrapper<Research>)techs.getSelectedItem()).getBase();
            rv = techtreeDependency.isEnabled( research );
        }
        return rv;
    }

    /**
     * @see sitool.components.ModulePane#update()
     */
    @Override
    public void update() {
        Account account = configData.getActiveAccount();
        if( currentAccount == null || !currentAccount.equals( account ) )
        {
            model.removeAll();
        }
        currentAccount = account;
        if( account != null )
        {
            center.setVisible( true );
            updateLabel();

            ItemEvent event;
            initTechs();
            event = new ItemEvent( techs, 0, techs.getSelectedItem(), ItemEvent.SELECTED );
            actionTechChanged( event );

            updateTable();
        }
        else
        {
            center.setVisible( false );
        }
    }

    private void updateLabel()
    {
        techLabel.setText( configData.getI18nText( "Research" ) );
        addButton.setText( configData.getI18nText( "add" ) );
        deleteButton.setText( configData.getI18nText( "delete" ) );
        deleteAll.setText( configData.getI18nText( "deleteAll" ) );
        importButton.setText( configData.getI18nText( "Load" ) );
        exportButton.setText( configData.getI18nText( "Save" ) );
    }

    private void updateTable()
    {
        model.fireTableStructureChanged();
        TableColumnModel columnModel = table.getColumnModel();
        TableColumn column;
        int width;
        column = table.getColumnModel().getColumn( TechBuilderTableModel.RESEARCH );
        width = 170;
        column.setPreferredWidth( width );
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column = columnModel.getColumn( TechBuilderTableModel.LEVEL );
        width = 36;
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column.setPreferredWidth( width );
        column = columnModel.getColumn( TechBuilderTableModel.TRANSMITTER );
        width = 55;
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column.setPreferredWidth( width );
        column = columnModel.getColumn( TechBuilderTableModel.POINTS );
        width = 65;
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column.setPreferredWidth( width );
        column = table.getColumnModel().getColumn( TechBuilderTableModel.DOWN );
        width = 16;
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column.setPreferredWidth( width );
        column = table.getColumnModel().getColumn( TechBuilderTableModel.UP );
        width = 16;
        column.setMinWidth( width );
        column.setMaxWidth( width );
        column.setPreferredWidth( width );
    }

    private void initTechs()
    {
        Object item = techs.getSelectedItem();
        techs.removeAllItems();
        for( Research research : Research.values() )
        {
            techs.addItem( new EnumWrapper<Research>( configData, research ) );
        }
        techs.setSelectedIndex(0);
        if(item != null) {
            techs.setSelectedItem(item);
        }
        techs.setMaximumRowCount( Research.values().length );
    }

    private Planet getResearchPlanet() {
        Planet researchPlanet = null;
        // check model, use that one first (planet is part of ListItem#equals()!)
        List<ListItem<Research>> data = model.getData();
        if( data.size() > 0) {
            Planet planet = data.get(0).getPlanet();
            researchPlanet = configData.getPlanet(planet);
        }
        else {
            // use the planet with the highest research lab
            int level = 0;
            for(Planet planet : currentAccount.getMyPlanets()) {
                Stage stage = planet.getStage(Building.researchLab, false);
                if(stage != null && stage.getLevel() > level) {
                    level = stage.getLevel();
                    researchPlanet = planet;
                }
            }
        }
        // update techtreeDependency
        if(researchPlanet != null) {
            techtreeDependency.setPlanets(Arrays.asList(researchPlanet));
        }
        return researchPlanet;
    }

    public Document getData(Parser xmlParser)
    {
        Document doc = null;
        if( xmlParser != null && currentAccount != null )
        {
            doc = xmlParser.getDocument();
            Element rootEl = doc.createElement( Parser.Tag_Account );
            doc.appendChild( rootEl );
            rootEl.setAttribute( Parser.Attribute_Name, currentAccount.getId() );
            int position = 0;
            for(ListItem<Research> item : model.getData() )
            {
                Element buildListNode = doc.createElement( Parser.Tag_TechlistItem );
                rootEl.appendChild( buildListNode );
                buildListNode.setAttribute( Parser.Attribute_Position, Integer.toString( position ) );
                buildListNode.setAttribute( Parser.Attribute_Tech, item.getItem().name() );
                Position planetPosition = item.getPlanet().getPosition();
                buildListNode.setAttribute( Parser.Attribute_Name, item.getPlanet().getName() );
                buildListNode.setAttribute( Parser.Attribute_Galaxy, Integer.toString( planetPosition.getGalaxy() ) );
                buildListNode.setAttribute( Parser.Attribute_System, Integer.toString( planetPosition.getSystem() ) );
                buildListNode.setAttribute( Parser.Attribute_Planet, Integer.toString( planetPosition.getPlanet() ) );
                buildListNode.setAttribute( Parser.Attribute_Level, Integer.toString( item.getLevel() ) );
                position++;
            }
        }
        return doc;
    }

    public String getExtension()
    {
        return extension;
    }

    public String getFileName()
    {
        return configData.getI18nText( "TechList.title" );
    }

    public MainFrame getMainFrame()
    {
        return mainFrame;
    }

    public File getCurrentDir()
    {
        return configData.getAccountDir();
    }

    public void setCurrentDir( File currentDir )
    {
        configData.setAccountDir( currentDir );
    }
}
